from typing import Tuple, Optional
import importlib

import moderngl
import pygame.image

from .sprites import SpriteArray, Sprite
from .atlas import Atlas
from .primitives.circles import Circle, line_vao, shape_vao
from .primitives.polygons import Polygon, Rect, PolyLine
from .primitives.text import Label, FontAtlas, text_vao
from .primitives.particles import ParticleGroup, ParticleVAO, PARTICLE_PROGRAM
from .loaders import images


class ShaderManager:
    def __init__(self, ctx):
        self.ctx = ctx
        self.programs = {}

    def get(self, vertex_shader, fragment_shader, geometry_shader=None):
        """Get a compiled program."""
        k = vertex_shader, fragment_shader
        try:
            return self.programs[k]
        except KeyError:
            pass

        prog = self.programs[k] = self.ctx.program(
            vertex_shader=vertex_shader,
            fragment_shader=fragment_shader,
            geometry_shader=geometry_shader,
        )
        return prog

    def set_proj(self, proj):
        """Set the projection matrix."""
        for prog in self.programs.values():
            try:
                uniform = prog['proj']
            except KeyError:
                continue
            uniform.write(proj.tobytes())


class FontManager:
    def __init__(self, ctx):
        self.ctx = ctx
        self.font_atlases = {}

    def get(self, font_name):
        try:
            return self.font_atlases[font_name]
        except KeyError:
            fa = self.font_atlases[font_name] = FontAtlas(self.ctx, font_name)
            return fa

    def update(self):
        """Send any dirty textures to the GL."""
        for f in self.font_atlases.values():
            f._update()


class Layer:
    def __init__(self, ctx, group):
        self.ctx = ctx
        self.group = group
        self.arrays = {}
        self.visible = True
        self.objects = set()
        self._dirty = set()
        self._dynamic = set()
        self.effect = None

    def clear(self):
        """Remove everything from the layer."""
        # We don't need to call .delete() on anything, faster to throw away
        # absolutely everything including VAOs.
        self.arrays.clear()
        self.objects.clear()
        self._dirty.clear()
        self._dynamic.clear()

    def render(self, t, dt):
        """Render the layer."""
        if not self.visible:
            return

        for o in self._dynamic:
            o._update(t, dt)

        for o in self._dirty:
            o._update()
        self._dirty.clear()

        if self.effect:
            self.effect.enter(t, dt)

        for a in self.arrays.values():
            a.render()

        if self.effect:
            self.effect.exit(t, dt)

    def set_effect(self, name, **kwargs):
        mod = importlib.import_module(f'wasabi2d.effects.{name}')
        cls = getattr(mod, name.title())
        self.effect = cls(self.ctx, self.group.shadermgr, **kwargs)
        self.effect._set_camera(self.group.camera)
        return self.effect

    def clear_effect(self):
        self.effect = None

    def add_sprite(self, image, pos=(0, 0), angle=0, anchor=None):
        spr = Sprite(
            layer=self,
            image=image,
            anchor=anchor
        )
        spr.pos = pos
        spr.angle = angle
        self.objects.add(spr)
        return spr

    def _migrate_sprite(self, spr, tex):
        """Move sprite spr into the correct vertex array."""
        k = ('sprite', tex.glo)
        array = self.arrays.get(k)
        if not array:
            prog = self.group.shadermgr.get(**SpriteArray.PROGRAM)
            array = SpriteArray(self.ctx, prog, tex, [spr])
            self.arrays[k] = array
        else:
            array.add(spr)

    def _get_or_create_vao(self, k, constructor):
        """Get a VAO identified by key k, or construct it using constructor."""
        vao = self.arrays.get(k)
        if not vao:
            vao = self.arrays[k] = constructor(self.ctx, self.group.shadermgr)
        return vao

    def _lines_vao(self):
        """Get a VAO for objects made of line strips."""
        return self._get_or_create_vao('lines', line_vao)

    def _fill_vao(self):
        """Get a VAO for objects made of colored triangles."""
        return self._get_or_create_vao('shapes', shape_vao)

    def _load_texture(self, name):
        """Load a texture."""
        img = images.load(name)
        data = pygame.image.tostring(img, "RGBA", 1)
        tex = self.ctx.texture(img.get_size(), 4, data=data)
        tex.build_mipmaps(max_level=2)
        return tex

    def _text_vao(self, font):
        """Get a VAO for objects made of font glyphs."""
        return self._get_or_create_vao(('text', font), text_vao)

    def add_circle(self,
                   *,
                   radius: float,
                   pos: Tuple[float, float] = (0, 0),
                   color: Tuple[float, float, float, float] = (1, 1, 1, 1),
                   fill: bool = True,
                   stroke_width: float = 1.0,
                   ) -> Circle:
        c = Circle(
            layer=self,
            radius=radius,
            pos=pos,
            color=color,
            stroke_width=stroke_width,
        )
        if fill:
            c._migrate_fill(self._fill_vao())
        else:
            c._migrate_stroke(self._lines_vao())
        self.objects.add(c)
        return c

    def add_star(
            self,
            *,
            points: int,
            inner_radius: float = 1.0,
            outer_radius: float = None,
            pos: Tuple[float, float] = (0, 0),
            fill: bool = True,
            color: Tuple[float, float, float, float] = (1, 1, 1, 1),
            stroke_width: float = 1.0,
    ) -> Circle:
        assert points >= 3, "Stars must have at least 3 points."

        if outer_radius is None:
            outer_radius = inner_radius * 2

        c = Circle(
            layer=self,
            segments=2 * points + 1,
            radius=1.0,
            pos=pos,
            color=color,
            stroke_width=stroke_width,
        )
        if fill:
            c._migrate_fill(self._fill_vao())
        else:
            c._migrate_stroke(self._lines_vao())
        self.objects.add(c)
        c.orig_verts[::2, :2] *= outer_radius
        c.orig_verts[1::2, :2] *= inner_radius
        return c

    def add_polygon(self,
                    vertices,
                    *,
                    pos: Tuple[float, float] = (0, 0),
                    color: Tuple[float, float, float, float] = (1, 1, 1, 1),
                    fill: bool = True,
                    stroke_width: float = 1.0,
                    ) -> Polygon:
        c = Polygon(
            layer=self,
            vertices=vertices,
            pos=pos,
            color=color,
            stroke_width=stroke_width,
        )
        self.objects.add(c)
        if fill:
            c._migrate_fill(self._fill_vao())
        else:
            c._migrate_stroke(self._lines_vao())
        return c

    def add_line(self,
                 vertices,
                 *,
                 pos: Tuple[float, float] = (0, 0),
                 color: Tuple[float, float, float, float] = (1, 1, 1, 1),
                 stroke_width: float = 1.0,
                 ) -> PolyLine:
        """Add a line strip.

        To create a single line segment of two points, pass two vertices!

        """
        c = PolyLine(
            layer=self,
            vertices=vertices,
            pos=pos,
            color=color,
            stroke_width=stroke_width,
        )
        self.objects.add(c)
        c._migrate_stroke(self._lines_vao())
        return c

    def add_rect(
            self,
            width: float,
            height: float,
            *,
            pos: Tuple[float, float] = (0, 0),
            color: Tuple[float, float, float, float] = (1, 1, 1, 1),
            fill: bool = True) -> Rect:

        c = Rect(
            layer=self,
            width=width,
            height=height,
            pos=pos,
            color=color
        )
        self.objects.add(c)
        if fill:
            c._migrate_fill(self._fill_vao())
        else:
            c._migrate_stroke(self._lines_vao())
        return c

    def add_label(self,
                  text: str,
                  *,
                  font: Optional[str] = None,
                  align: str = 'left',
                  fontsize: int = 20,
                  pos: Tuple[float, float] = (0, 0),
                  color: Tuple[float, float, float, float] = (1, 1, 1, 1),
                  ) -> Rect:

        fa = self.group.fontmgr.get(font)
        c = Label(
            text,
            fa,
            self,
            align=align,
            fontsize=fontsize,
            pos=pos,
            color=color
        )
        self.objects.add(c)
        c._migrate(self._text_vao(font))
        return c

    def add_particle_group(self, texture=None, **kwargs) -> ParticleGroup:
        """Create a group of particles.

        We do not actually emit any particles at this time.
        """
        c = ParticleGroup(layer=self, **kwargs)

        vao = self.arrays[c] = ParticleVAO(
            c,
            mode=moderngl.POINTS,
            ctx=self.ctx,
            prog=self.group.shadermgr.get(**PARTICLE_PROGRAM),
        )
        if texture is None:
            tex = self.ctx.texture((1, 1), 4, data=b'\xff' * 4)
        else:
            tex = self._load_texture(texture)
        vao.tex = tex
        vao.color_tex = c.color_tex

        self.objects.add(c)
        self._dynamic.add(c)
        c._migrate(vao)
        return c


class LayerGroup(dict):
    def __new__(cls, ctx, camera):
        return dict.__new__(cls)

    def __init__(self, ctx, camera):
        self.ctx = ctx
        self.camera = camera
        self.shadermgr = ShaderManager(self.ctx)
        self.fontmgr = FontManager(self.ctx)
        self.atlas = Atlas(ctx)

    def __missing__(self, k):
        if not isinstance(k, (float, int)):
            raise TypeError("Layer indices must be numbers")
        layer = self[k] = Layer(self.ctx, self)
        return layer

    def render(self, proj, t, dt):
        self.atlas._update()
        self.fontmgr.update()
        self.shadermgr.set_proj(proj)
        for k in sorted(self):
            self[k].render(t, dt)



import random
import colorsys
from collections import deque
from math import copysign
from wasabi2d import event, run, Scene, animate, keys
from wasabi2d.actor import Actor
import pygame.mouse


WIDTH = 600
HEIGHT = 800
scene = Scene(WIDTH, HEIGHT)
pygame.mouse.set_visible(False)

BALL_SIZE = 6
MARGIN = 50

BRICKS_X = 10
BRICKS_Y = 5
BRICK_W = (WIDTH - 2 * MARGIN) / BRICKS_X
BRICK_H = 25


ball = Actor(
    scene.layers[1].add_circle(
        radius=BALL_SIZE,
        color='#cccccc',
    ),
    pos=(WIDTH / 2, HEIGHT / 2),
)
scene.layers[1].set_effect('trails', fade=0.2)
bat = Actor(
    scene.layers[0].add_rect(
        width=120,
        height=12,
        color='pink',
        fill=True,
    ),
    pos=(WIDTH / 2, HEIGHT - 50)
)


bricks = []


def reset():
    """Reset bricks and ball."""
    # First, let's do bricks
    for b in bricks:
        b.delete()
    del bricks[:]
    for x in range(BRICKS_X):
        for y in range(BRICKS_Y):
            hue = (x + y) / BRICKS_X
            saturation = (y / BRICKS_Y) * 0.5 + 0.5
            rect = scene.layers[0].add_rect(
                color=colorsys.hsv_to_rgb(hue, saturation, 0.8),
                width=BRICK_W,
                height=BRICK_H,
                fill=True,
            )
            brick = Actor(
                rect,
                pos=((x + 0.5) * BRICK_W + MARGIN,
                     (y + 0.5) * BRICK_H + MARGIN),
            )
            #brick.highlight = hsv_color(hue, saturation * 0.7, 1.0)
            bricks.append(brick)

    # Now reset the ball
    ball.pos = (WIDTH / 2, HEIGHT / 3)
    ball.vel = (random.uniform(-200, 200), 400)


# Reset bricks and ball at start
reset()


@event
def update():
    # When you have fast moving objects, like the ball, a good trick
    # is to run the update step several times per frame with tiny time steps.
    # This makes it more likely that collisions will be handled correctly.
    for _ in range(3):
        update_step(1 / 180)
    update_bat_vx()


def update_step(dt):
    x, y = ball.pos
    vx, vy = ball.vel

    if ball.top > HEIGHT:
        reset()
        return

    # Update ball based on previous velocity
    x += vx * dt
    y += vy * dt
    ball.pos = (x, y)

    # Check for and resolve collisions
    if ball.left < 0:
        vx = abs(vx)
        ball.left = -ball.left
    elif ball.right > WIDTH:
        vx = -abs(vx)
        ball.right -= 2 * (ball.right - WIDTH)

    if ball.top < 0:
        vy = abs(vy)
        ball.top *= -1

    if ball.colliderect(bat):
        vy = -abs(vy)
        # Add some spin off the paddle
        vx += -30 * bat.vx
    else:
        # Find first collision
        idx = ball.collidelist(bricks)
        if idx != -1:
            scene.camera.screen_shake()
            brick = bricks[idx]

            # Work out what side we collided on
            dx = (ball.centerx - brick.centerx) / BRICK_W
            dy = (ball.centery - brick.centery) / BRICK_H
            if abs(dx) > abs(dy):
                vx = copysign(abs(vx), dx)
            else:
                vy = copysign(abs(vy), dy)
            del bricks[idx]

            rect = brick.prim
            animate(
                rect,
                tween='bounce_end',
                scale=0
            )
            animate(
                rect,
                on_finished=rect.delete,
                color=(0, 0, 0, 1),
                angle=random.uniform(-1, 1),
            )

    ball.vel = (vx, vy)


# Keep bat vx history over 5 frames
bat.recent_vxs = deque(maxlen=5)
bat.vx = 0
bat.prev_centerx = bat.pos[0]


def update_bat_vx():
    """Recalculate average bat vx."""
    x = bat.pos[0]
    dx = x - bat.prev_centerx
    bat.prev_centerx = x

    history = bat.recent_vxs
    history.append(dx)
    vx = sum(history) / len(history)
    bat.vx = min(10, max(-10, vx))


@event
def on_mouse_move(pos):
    x, y = pos
    bat.pos = x, bat.pos[1]
    if bat.left < 0:
        bat.left = 0
    elif bat.right > WIDTH:
        bat.right = WIDTH


@event
def on_key_down(key):
    if key == keys.F12:
        scene.toggle_recording()


run()



import sys
import operator
import time
import types
from collections import namedtuple

import pygame

import wasabi2d.clock
import wasabi2d.keyboard
import wasabi2d.loaders
from . import constants


screen = None  # This global surface is what actors draw to
DISPLAY_FLAGS = 0


def exit():
    """Wait for up to a second for all sounds to play out
    and then exit
    """
    t0 = time.time()
    while pygame.mixer.get_busy():
        time.sleep(0.1)
        if time.time() - t0 > 1.0:
            break
    sys.exit()


def positional_parameters(handler):
    """Get the positional parameters of the given function."""
    code = handler.__code__
    return code.co_varnames[:code.co_argcount]


class DEFAULTICON:
    """Sentinel indicating that we want to use the default icon."""


DrawEvent = namedtuple('DrawEvent', 'type t dt')
UpdateEvent = namedtuple('UpdateEvent', 'type t dt keyboard')


class EventMapper:
    def __init__(self):
        self.keyboard = wasabi2d.keyboard.keyboard
        self.handlers = {}
        self.lock_fps = False

    EVENT_HANDLERS = {
        'on_mouse_down': pygame.MOUSEBUTTONDOWN,
        'on_mouse_up': pygame.MOUSEBUTTONUP,
        'on_mouse_move': pygame.MOUSEMOTION,
        'on_key_down': pygame.KEYDOWN,
        'on_key_up': pygame.KEYUP,
        'on_music_end': constants.MUSIC_END,
        'draw': DrawEvent,
        'update': UpdateEvent,
    }

    def map_buttons(val):
        return {c for c, pressed in zip(constants.mouse, val) if pressed}

    EVENT_PARAM_MAPPERS = {
        'buttons': map_buttons,
        'button': constants.mouse,
        'key': constants.keys
    }

    def __call__(self, handler):
        """Register an event handler."""
        name = handler.__name__
        try:
            type = self.EVENT_HANDLERS[name]
        except KeyError:
            raise KeyError(
                f"Unknown handler type {name}"
            )
        self.handlers[type] = self.prepare_handler(handler)
        return handler

    def prepare_handler(self, handler):
        """Adapt a wasabi2d game's raw handler function to take a Pygame Event.

        Returns a one-argument function of the form ``handler(event)``.
        This will ensure that the correct arguments are passed to the raw
        handler based on its argument spec.

        The wrapped handler will also map certain parameter values using
        callables from EVENT_PARAM_MAPPERS; this ensures that the value of
        'button' inside the handler is a real instance of constants.mouse,
        which means (among other things) that it will print as a symbolic value
        rather than a naive integer.

        """
        code = handler.__code__
        if isinstance(handler, types.MethodType):
            start_idx = 1
        else:
            start_idx = 0

        param_names = code.co_varnames[start_idx:code.co_argcount]

        def make_getter(mapper, getter):
            if mapper:
                return lambda event: mapper(getter(event))
            return getter

        param_handlers = []
        for name in param_names:
            getter = operator.attrgetter(name)
            mapper = self.EVENT_PARAM_MAPPERS.get(name)
            param_handlers.append((name, make_getter(mapper, getter)))

        def prep_args(event):
            return {name: get(event) for name, get in param_handlers}

        def new_handler(event):
            try:
                prepped = prep_args(event)
            except ValueError:
                # If we couldn't construct the keys/mouse objects representing
                # the button that was pressed, then skip the event handler.
                #
                # This happens because Pygame can generate key codes that it
                # does not have constants for.
                return
            else:
                return handler(**prepped)

        return new_handler

    def dispatch_event(self, event):
        if event.type == pygame.QUIT:
            sys.exit(0)
        if event.type == pygame.KEYDOWN:
            if event.key == pygame.K_q and \
                    event.mod & (pygame.KMOD_CTRL | pygame.KMOD_META):
                sys.exit(0)
            self.keyboard._press(event.key)
        elif event.type == pygame.KEYUP:
            self.keyboard._release(event.key)

        handler = self.handlers.get(event.type)
        if handler:
            self.need_redraw = True
            handler(event)
            return True
        return False

    def run(self):
        """Run the main loop."""
        clock = pygame.time.Clock()
        # self.reinit_screen()

        pgzclock = wasabi2d.clock.clock

        self.need_redraw = True
        t = 0
        while True:
            dt = clock.tick(60) / 1000.0

            if self.lock_fps:
                dt = 1.0 / 60.0

            t += dt

            for event in pygame.event.get():
                self.dispatch_event(event)

            pgzclock.tick(dt)

            ev = UpdateEvent(UpdateEvent, t, dt, self.keyboard)
            updated = self.dispatch_event(ev)

            screen_change = False  # self.reinit_screen()
            if screen_change or updated or pgzclock.fired or self.need_redraw:
                self.dispatch_event(DrawEvent(DrawEvent, t, dt))
                pygame.display.flip()
                self.need_redraw = False



"""Pi Lander

 * A basic Lunar Lander style game in Pygame Zero
 * Run with 'pgzrun pi_lander.py', control with the LEFT, RIGHT and UP arrow
   keys
 * Original Author Tim Martin: www.Tim-Martin.co.uk
 * Ported to Wasabi2D by Daniel Pope
 * Licence: Creative Commons Attribution-ShareAlike 4.0 International
 * http://creativecommons.org/licenses/by-sa/4.0/

"""
import random
import math
from wasabi2d import Scene, event, run, clock, keys, Vector2


WIDTH = 1600  # Screen width
HEIGHT = 1200  # Screen height

# Landscape is broken down into steps. Define number of pixels on the x axis
# per step.
STEP_SIZE = 12

scene = Scene(
    WIDTH,
    HEIGHT,
    antialias=4,
    title="Lunar Lander",
)


class LandingSpot:
    """A flat pad where it is safe for the player to land.

    Each instance defines a landing spot by where it starts, how big it is and
    how many points it's worth.
    """

    landing_spot_sizes = ["small", "medium", "large"]

    def __init__(self, starting_step):
        self.starting = starting_step
        random_size = random.choice(
            LandingSpot.landing_spot_sizes
        )  # And randomly choose size
        if random_size == "small":
            self.size = 4
            self.bonus = 8
        elif random_size == "medium":
            self.size = 10
            self.bonus = 4
        else:  # Large
            self.size = 20
            self.bonus = 2

        self.px_width = (self.size - 1) * STEP_SIZE
        self.px_x = self.starting * STEP_SIZE + self.px_width / 2
        self.px_y = 0  # updated when we generate the landscape

    def get_within_landing_spot(self, step):
        if (step >= self.starting) and (step < self.starting + self.size):
            return True
        return False


class Landscape:
    """ Stores and generates the landscape, landing spots and star field """


    world_steps = int(
        WIDTH / STEP_SIZE
    )  # How many steps can we fit horizontally on the screen
    SMALL_HEIGHT_CHANGE = 6  # Controls how bumpy the landscape is
    LARGE_HEIGHT_CHANGE = 20  # Controls how steep the landscape is
    FEATURES = ["mountain", "valley", "field"]  # What features to generate
    N_STARS = 100  # How many stars to put in the background
    n_spots = 4  # Max number of landing spots to generate

    # the boundary of starting landscape height, dependent on screen height
    MIN_START_Y = int(0.5 * HEIGHT)
    MAX_START_Y = int(0.83 * HEIGHT)

    MOUNTAIN_START_THRESHOLD = int(0.95 * HEIGHT)
    VALLEY_START_THRESHOLD = int(0.33 * HEIGHT)

    def __init__(self):
        self.world_height = []  # Holds the height of the landscape at each step
        self.star_locations = []  # Holds the x and y location of the stars
        self.landing_spots = []  # Holds the landing spots

    def get_landing_spot(self, step):
        """Get the landing spot for this step."""
        for spot in self.landing_spots:
            if spot.get_within_landing_spot(step):
                return spot
        return None

    def get_within_landing_spot(self, step):
        """ Calculate if a given step is within any of the landing spots """
        return self.get_landing_spot(step) is not None

    def get_landing_spot_bonus(self, step):
        """Get the bonus if we're at a landing spot."""
        spot = self.get_landing_spot(step)
        return spot.bonus if spot else 0

    def reset(self):
        """Generates a new landscape."""
        scene.layers[-4].clear()
        scene.layers[-3].clear()
        scene.layers[-2].clear()
        scene.layers[-1].clear()
        self.generate()

    def generate(self):
        # First: Choose which steps of the landscape will be landing spots
        del self.landing_spots[:]  # Delete any previous LandingSpot objects
        next_spot_start = 0
        # Move from left to right adding new landing spots until either
        # n_spots spots have been placed or we run out of space in the world
        while (
            len(self.landing_spots) < Landscape.n_spots
            and next_spot_start < Landscape.world_steps
        ):

            # Randomly choose location to start landing spot
            next_spot_start += random.randint(10, 50)
            # Make a new landing object at this spot
            new_landing_spot = LandingSpot(next_spot_start)
            # And store it in our list
            self.landing_spots.append(new_landing_spot)

            # Then take into account its size before choosing the next
            next_spot_start += new_landing_spot.size

        # Second: Randomise the world map
        del self.world_height[:]  # Clear any previous world height data
        feature_steps = 0  # Keep track of how many steps we are into a feature

        # Start the landscape between 300 and 500 pixels down
        self.world_height.append(random.randint(Landscape.MIN_START_Y, Landscape.MAX_START_Y))
        for step in range(1, Landscape.world_steps):
            # If feature_step is zero, we need to choose a new feature and how long it goes on for
            if feature_steps == 0:
                feature_steps = random.randint(25, 75)
                current_feature = random.choice(self.FEATURES)

            current_height = self.world_height[step - 1]
            # Generate the world by setting the range of random numbers, must be flat if in a landing spot
            spot = self.get_landing_spot(step)
            if spot is not None:
                spot.px_y = current_height
                max_up = 0  # Flat
                max_down = 0  # Flat
            elif current_feature == "mountain":
                max_up = Landscape.SMALL_HEIGHT_CHANGE
                max_down = -Landscape.LARGE_HEIGHT_CHANGE
            elif current_feature == "valley":
                max_up = Landscape.LARGE_HEIGHT_CHANGE
                max_down = -Landscape.SMALL_HEIGHT_CHANGE
            elif current_feature == "field":
                max_up = Landscape.SMALL_HEIGHT_CHANGE
                max_down = -Landscape.SMALL_HEIGHT_CHANGE

            # Generate the next piece of the landscape
            next_height = current_height + random.randint(max_down, max_up)
            self.world_height.append(next_height)
            feature_steps -= 1
            # Stop mountains getting too high, or valleys too low
            if next_height > Landscape.MOUNTAIN_START_THRESHOLD:
                current_feature = "mountain"  # Too low! Force a mountain
            elif next_height < Landscape.VALLEY_START_THRESHOLD:
                current_feature = "valley"  # Too high! Force a valley

        self.horizon = scene.layers[-2].add_line(
            self.points(),
            stroke_width=3,
            color='#555555'
        )
        for spot in self.landing_spots:
            scene.layers[-1].add_rect(
                pos=(spot.px_x, spot.px_y),
                width=spot.px_width,
                height=8,
                color='yellow'
            )
            scene.layers[-3].add_label(
                f'{spot.bonus}x',
                align='center',
                pos=(spot.px_x, spot.px_y - 20),
                fontsize=40,
                color="#888800",
            )

        # Third: Randomise the star field
        del self.star_locations[:]
        for star in range(0, Landscape.N_STARS):
            star_step = random.randint(0, Landscape.world_steps - 1)
            star_x = star_step * STEP_SIZE
            star_y = random.randint(
                0, self.world_height[star_step]
            )  # Keep the stars above the landscape

            mag = random.random()
            star = scene.layers[-4].add_star(
                points=4,
                inner_radius=2,
                outer_radius=10,
                color=(mag,) * 3,
                pos=(star_x, star_y),
            )
            star.scale = mag
            star.angle = math.radians(45)

    def points(self):
        points = []
        for step in range(0, self.world_steps):
            x = step * STEP_SIZE
            y = self.world_height[step]
            points.append((x, y))
        return points


class Ship:
    """ Holds the state of the player's ship and handles movement """

    max_fuel = 1000  # How much fuel the player starts with
    booster_power = 0.1  # Power of the ship's thrusters
    rotate_speed = 3  # How fast the ship rotates in degrees per frame
    gravity = [0.0, 0.02]  # Strength of gravity in the x and y directions

    def __init__(self):
        """ Create the variables which will describe the players ship """
        self.angle = 0  # The angle the ship is facing 0 - 360 degrees
        self.altitude = 0  # The number of pixels the ship is above the ground
        self.booster = False  # True if the player is firing their booster
        self.fuel = 0  # Amount of fuel remaining
        self.position = [0, 0]  # The x and y coordinates of the players ship
        self.velocity = [0, 0]  # The x and y velocity of the players ship
        self.acceleration = [0, 0]  # The x and y acceleration of the players ship
        self.sprite = scene.layers[0].add_sprite('lander')
        self.particles = scene.layers[0].add_particle_group(
            max_age=1,
            grow=10,
        )
        self.particles.add_color_stop(0, (5, 0, 0, 1))
        self.particles.add_color_stop(0.2, (1.5, 1.5, 0, 1))
        self.particles.add_color_stop(0.5, 'gray')
        self.particles.add_color_stop(0.75, (5, 0, 0, 0))
        scene.layers[0].set_effect('bloom', radius=4)

    def reset(self):
        """ Set the ships position, velocity and angle to their new-game values """
        self.position = [750.0, 100.0]  # Always start at the same spot
        self.sprite.pos = self.position
        self.velocity = [
            -random.random(),
            random.random(),
        ]  # But with some initial speed
        self.acceleration = [
            0.0,
            0.0,
        ]  # No initial acceleration (except gravity of course)
        self.angle = random.randint(0, 360)  # And pointing in a random direction
        self.sprite.angle = math.radians(-self.angle)
        self.fuel = Ship.max_fuel  # Fill up fuel tanks

    def rotate(self, direction):
        """ Rotate the players ship and keep the angle within the range 0 - 360 degrees """
        if direction == "left":
            self.angle -= Ship.rotate_speed
        elif direction == "right":
            self.angle += Ship.rotate_speed
        if (
            self.angle > 360
        ):  # Remember than adding or subtracting 360 degrees does not change the angle
            self.angle -= 360
        elif self.angle < 0:
            self.angle += 360
        self.sprite.angle = math.radians(-self.angle)

    def booster_on(self):
        """ When booster is firing we accelerate in the opposite direction, 180 degrees, from the way the ship is facing """
        self.booster = True
        self.sprite.image = 'lander-thrust'

        angle_r = math.radians(self.angle + 180)

        up = Vector2(math.sin(angle_r), math.cos(angle_r))

        self.acceleration[:] = Ship.booster_power * up

        self.particles.emit(
            15,
            pos=self.position - 25 * up,
            pos_spread=2,
            vel=up * -200 + Vector2(*self.velocity) * 60,
            vel_spread=50,
            spin_spread=1,
            size=1.5,
            #color='#fff0c0',
        )

        self.fuel -= 2

    def booster_off(self):
        """ When the booster is not firing we do not accelerate """
        self.sprite.image = 'lander'
        self.booster = False
        self.acceleration[0] = 0.0
        self.acceleration[1] = 0.0

    def update_physics(self):
        """ Update ship physics in X and Y, apply acceleration (and gravity) to the velocity and velocity to the position """
        for axis in range(0, 2):
            self.velocity[axis] += Ship.gravity[axis]
            self.velocity[axis] += self.acceleration[axis]
            self.position[axis] += self.velocity[axis]

        # Update player altitude. Note that STEP_SIZE * 3 is the length of the
        # ship's legs
        ship_step = int(self.position[0] / STEP_SIZE)
        if ship_step < Landscape.world_steps:
            self.altitude = (
                game.landscape.world_height[ship_step]
                - self.position[1]
                - (STEP_SIZE * 3)
            )

        self.sprite.pos = self.position

    def get_out_of_bounds(self):
        """ Check if the player has hit the ground or gone off the sides """
        return (
            self.altitude <= 0
            or self.position[0] <= 0
            or self.position[0] >= WIDTH
        )


class Game:
    """ Holds main game data, including the ship and landscape objects. Checks for game-over """

    def __init__(self):
        self.time = 0.0  # Time spent playing in seconds
        self.score = 0  # Player's score
        self.game_speed = 30  # How fast the game should run in frames per second
        self.time_elapsed = 0.0  # Time since the last frame was changed
        self.blink = True  # True if blinking text is to be shown
        self.n_frames = 0  # Number of frames processed
        self.game_on = False  # True if the game is being played
        game_label.text = "PI   LANDER\nPRESS SPACE TO START"  # Start of game message
        self.ship = Ship()  # Make a object of the Ship type
        self.landscape = Landscape()
        self.reset()  # Start the game with a fresh landscape and ship

    def reset(self):
        self.time = 0.0
        self.ship.reset()
        self.landscape.reset()

    def restart(self):
        self.reset()
        scene.layers[5].visible = False
        self.game_on = True

    def end_game(self, message):
        game_label.text = message
        scene.layers[5].visible = True
        self.game_on = False

    def check_game_over(self):
        """ Check if the game is over and update the game state if so """
        if self.ship.get_out_of_bounds() == False:
            return  # Game is not over

        self.game_on = False  # Game has finished. But did we win or loose?
        # Check if the player looses. This is if the ship's angle is > 20 degrees
        # the ship is not over a landing site, is moving too fast or is off the side of the screen
        ship_step = int(self.ship.position[0] / STEP_SIZE)
        if (
            self.ship.position[0] <= 0
            or self.ship.position[0] >= WIDTH
            or self.landscape.get_within_landing_spot(ship_step) == False
            or abs(self.ship.velocity[0]) > 1
            or abs(self.ship.velocity[1]) > 1
            or (self.ship.angle > 20 and self.ship.angle < 340)
        ):
            self.end_game(
                "YOU JUST DESTROYED A 100 MEGABUCK LANDER\n\n"
                "LOSE 250 POINTS\n\n"
                "PRESS SPACE TO RESTART"
            )
            self.score -= 250
        else:  # If the player has won! Update their score based on the amount of remaining fuel and the landing bonus
            points = self.ship.fuel / 10
            points *= self.landscape.get_landing_spot_bonus(ship_step)
            self.score += points
            self.end_game(
                "CONGRATULATIONS\nTHAT WAS A GREAT LANDING!\n\n"
                f"{round(points)} POINTS\n\nPRESS SPACE TO RESTART"
            )


game_label = scene.layers[5].add_label(
    "Press space to play",
    font='eunomia_regular',
    pos=(WIDTH / 2, HEIGHT / 5),
    align="center",
    color="green",
    fontsize=60,
)

# Create the game object
game = Game()


def toggle_spots():
    if game.game_on:
        scene.layers[-3].visible = True
    else:
        scene.layers[-3].visible ^= 1


clock.schedule_interval(toggle_spots, 0.5)

hud = scene.layers[2]

score_label = hud.add_label(
    "SCORE: " + str(round(game.score)),
    font='eunomia_regular',
    pos=(10, 55),
    fontsize=40,
)
time_label = hud.add_label(
    "TIME: " + str(round(game.time)),
    font='eunomia_regular',
    pos=(10, 95),
    fontsize=40,
)
fuel_label = hud.add_label(
    "FUEL: " + str(game.ship.fuel),
    font='eunomia_regular',
    pos=(10, 135),
    fontsize=40,
)
alt_label = hud.add_label(
    "ALTITUDE: " + str(round(game.ship.altitude)),
    font='eunomia_regular',
    pos=(WIDTH - 10, 55),
    align='right',
    fontsize=40,
)
vx_label = hud.add_label(
    "HORIZONTAL SPEED: {0:.2f}".format(game.ship.velocity[0]),
    font='eunomia_regular',
    pos=(WIDTH - 10, 95),
    align='right',
    fontsize=40,
)
vy_label = hud.add_label(
    "VERTICAL SPEED: {0:.2f}".format(-game.ship.velocity[1]),
    font='eunomia_regular',
    pos=(WIDTH - 10, 135),
    align='right',
    fontsize=40,
)


@event
def update(dt, keyboard):
    update_hud()
    update_physics(dt, keyboard)


def update_hud():
    score_label.text = f"SCORE: {round(game.score)}"
    time_label.text = f"TIME: {round(game.time)}"
    fuel_label.text = f"FUEL: {game.ship.fuel}"
    alt_label.text = f"ALTITUDE: {round(game.ship.altitude)}"

    vx, vy = game.ship.velocity
    vx_label.text = f"HORIZONTAL SPEED: {vx:.2f}"
    vy_label.text = f"VERTICAL SPEED: {-vy:.2f}"


def update_physics(dt, keyboard):
    """Updates the game physics 30 times every second."""
    game.time_elapsed += dt

    # New frame - do all the simulations
    game.n_frames += 1

    # Start the game if the player presses space when the game is not on
    if game.game_on == False:
        return

    # If the game is on, update the movement and the physics
    if keyboard.left:  # Change space ship rotation
        game.ship.rotate("left")
    elif keyboard.right:
        game.ship.rotate("right")

    if (
        keyboard.up and game.ship.fuel > 0
    ):  # Fire boosters if the player has enough fuel
        game.ship.booster_on()
    else:
        game.ship.booster_off()

    game.time += dt
    game.ship.update_physics()
    game.check_game_over()


@event
def on_key_down(key):
    if key == keys.F12:
        scene.toggle_recording()
    elif key == keys.SPACE and game.game_on == False:
        game.restart()


run()



from typing import Union

from sortedcontainers import SortedList


class NoCapacity(Exception):
    """Raised when the allocator is out of capacity to allocate a block.

    The attribute `.recommended` will contain a recommended size to grow to.
    """


class AbstractAllocator:
    """Manage allocations within a block of items.

    This is abstract because we don't actually assume anything about the types
    of the items or how to write to them.

    """

    def __init__(self, capacity: int = 8192):
        self.capacity = capacity
        self.allocs = {}
        self._free = SortedList([(capacity, 0)])

    def avail(self) -> int:
        """Get the current available capacity."""
        return sum(cap for cap, off in self._free)

    def grow(self, new_capacity: int):
        """Grow the available space for allocation to new_capacity.

        No reallocations or compactions are done here so the only guaranteed
        contiguous new free space is at the end.
        """
        self._release(self.capacity, new_capacity - self.capacity)
        self.capacity = new_capacity

    def _release(self, offset, length):
        """Release the given block back to the pool.

        We consider joining this block to a subsequent one. We don't currently
        consider joining it to the previous block.

        """
        if not length:
            return

        while True:
            k = (length, offset)
            idx = self._free.bisect_left(k)
            if idx >= len(self._free):
                self._free.add(k)
                return

            next_length, next_off = self._free[idx]
            if next_off != (length + offset):
                return

            self._free.pop(idx)
            length += next_length

    def alloc(self, num: int) -> slice:
        """Allocate a block of size num.

        Return a slice, or raise NoCapacity if there was insufficient
        capacity available.

        """
        pos = self._free.bisect_left((num, 0))
        if pos == len(self._free):
            # capacity is not high enough
            new_capacity = self.capacity
            while new_capacity < num:
                new_capacity *= 2

            err = NoCapacity()
            err.recommended = self.capacity + new_capacity
            raise err

        # We have located where we want to reserve, insert it
        return self._reserve(pos, num)

    def _reserve(self, pos, num) -> slice:
        """Update the free list with the given reservation.

        Return the reserved slice.
        """
        block_size, offset = self._free.pop(pos)

        # Release the rest of the block in power-of-2 blocks
        mid = block_size // 2
        while mid >= num >= 2:
            self._release(offset + mid, block_size - mid)
            block_size = mid
            mid = block_size // 2

        end_off = offset + num

        self._release(end_off, block_size - num)

        # Store the size of the block that we allocated
        self.allocs[offset] = num
        return slice(offset, end_off)

    def realloc(self, offset: Union[int, slice], new_size: int) -> slice:
        """Reallocate the given block.

        This is optimised so that if there is extra space in the original
        allocation, it can be done without moving the block.

        This operation can fail, raising NoCapacity.

        """
        if isinstance(offset, slice):
            offset = offset.start

        try:
            size = self.allocs[offset]
        except KeyError:
            raise KeyError(f"Offset {offset} is not allocated.") from None
        if new_size <= size:
            return slice(offset, offset + new_size)

        del self.allocs[offset]
        self._release(offset, size)
        try:
            return self.alloc(new_size)
        except NoCapacity:
            # We don't have enough capacity, re-insert before raising
            self._reserve(offset, size)
            raise

    def free(self, offset: Union[int, slice]):
        """Free the block at offset."""
        if isinstance(offset, slice):
            offset = offset.start
        try:
            size = self.allocs.pop(offset)
        except KeyError:
            raise KeyError(f"Offset {offset} is not allocated.") from None
        self._release(offset, size)



from typing import Tuple, Any

import numpy as np
import numpy.random
from sortedcontainers import SortedList

from ..color import convert_color
from ..allocators.vertlists import VAO


PARTICLE_PROGRAM = dict(
    vertex_shader='''
        #version 330

        in vec2 in_vert;
        in vec4 in_color;
        in float in_size;
        in float in_angle;
        in float in_age;
        out vec4 g_color;
        out float size;
        out float age;
        out mat2 rots;

        uniform float max_age;

        void main() {
            gl_Position = vec4(in_vert, 0.0, 1.0);
            g_color = in_color;
            size = in_size;
            age = in_age;

            float c = cos(in_angle);
            float s = sin(in_angle);
            rots = mat2(c, -s, s, c);
        }
    ''',
    geometry_shader="""
#version 330 core
layout (points) in;
layout (triangle_strip, max_vertices = 4) out;

in mat2 rots[];
in vec4 g_color[];
in float size[];
in float age[];
out vec4 color;
out vec2 uv;

uniform mat4 proj;
uniform sampler2D color_tex;
uniform float max_age;
uniform float grow;

void main() {
    float age_frac = clamp(age[0] / max_age, 0.0, 511.0 / 512.0);
    color = g_color[0] * texture(color_tex, vec2(age_frac, 0.0));
    vec2 pos = gl_in[0].gl_Position.xy;
    mat2 rot = rots[0];

    float sz = size[0] * pow(grow, age[0]);

    // Vector to the corner
    vec2 corners[4] = vec2[4](
        vec2(-sz, sz),
        vec2(sz, sz),
        vec2(-sz, -sz),
        vec2(sz, -sz)
    );
    vec2 uvs[4] = vec2[4](
        vec2(0, 1),
        vec2(1, 1),
        vec2(0, 0),
        vec2(1, 0)
    );

    for (int i = 0; i < 4; i++) {
        gl_Position = proj * vec4(pos + rot * corners[i], 0.0, 1.0);
        uv = uvs[i];
        EmitVertex();
    }
    EndPrimitive();
}
""",
    fragment_shader='''
        #version 330

        out vec4 f_color;
        in vec4 color;
        in vec2 uv;
        uniform sampler2D tex;

        void main() {
            f_color = color * texture(tex, uv);
        }
    ''',
)

PARTICLE_DTYPE = np.dtype([
    ('in_vert', '2f4'),
    ('in_color', '4f4'),
    ('in_age', 'f4'),
    ('in_size', 'f4'),
    ('in_angle', 'f4'),
])


class ParticleVAO(VAO):
    """A VAO with an image texture and a color ramp texture."""

    def __init__(self, pgroup, *args, **kwargs):
        super().__init__(*args, dtype=PARTICLE_DTYPE, **kwargs)
        self.pgroup = pgroup

    def render(self):
        self.prog['grow'].value = self.pgroup.grow
        self.prog['max_age'].value = self.pgroup.max_age
        self.prog['tex'].value = 0
        self.prog['color_tex'].value = 1
        self.tex.use(0)
        self.color_tex.use(1)
        super().render()


class ParticleGroup:
    """A group of particles."""

    def __init__(
            self,
            layer,
            *,
            grow: float = 1.0,
            max_age: float = np.inf,
            gravity: Tuple[float, float] = (0, 0),
            drag: float = 1.0):
        super().__init__()
        self.layer = layer
        self.num: int = 0  # Number of particles we have
        self.max_age = max_age
        self.grow = grow
        self.gravity = np.array(gravity)
        self.drag = drag
        self.spins = np.zeros([0])
        self.vels = np.zeros([0, 2])
        self._color_stops = SortedList()
        self.color_tex = layer.ctx.texture((512, 1), 4, dtype='f2')
        self.color_vals = np.ones((512, 4), dtype='f2')
        self.color_tex.write(self.color_vals)

    def add_color_stop(self, age, color):
        """Add a color stop for particles of the given age.

        Particles will fade between the colors of the stops as their age
        changes.
        """
        color = convert_color(color)
        self._color_stops.add((age, *color))
        xs = np.linspace(0, self.max_age, 512)

        ages, *colors = zip(*self._color_stops)
        for i in range(4):
            self.color_vals[:, i] = np.interp(
                xs,
                ages,
                colors[i],
            )
        self.color_tex.write(self.color_vals)

    def emit(
            self,
            num: int,
            *,
            pos: Tuple[float, float],
            pos_spread: float = 0,
            vel: Tuple[float, float] = (0, 0),
            vel_spread: float = 0,
            color: Any = (1, 1, 1, 1),
            size: float = 1.0,
            size_spread: float = 0.0,
            spin: float = 0.0,
            spin_spread: float = 0.0,
            ):
        """Emit num particles."""
        num = round(num)
        color = convert_color(color)

        prev_verts = self.lst.vertbuf
        alive = self.lst.vertbuf['in_age'] < self.max_age
        num_alive = np.sum(alive)
        need = num_alive + num

        verts_alive = prev_verts[alive]

        self.lst.realloc(need, need)
        self.lst.indexbuf[:] = np.arange(need, dtype='u4') + self.lst.vertoff.start

        new_vel = np.random.normal(vel, vel_spread, [num, 2])
        new_pos = np.random.normal(pos, pos_spread, [num, 2])
        new_size = np.random.normal(size, size_spread, num)
        new_spins = np.random.normal(spin, spin_spread, num)

        verts = self.lst.vertbuf
        self.vels = np.vstack([self.vels[alive], new_vel])
        self.spins = np.hstack([self.spins[alive], new_spins])
        verts[:num_alive] = verts_alive
        verts[num_alive:]['in_age'] = 0
        verts[num_alive:]['in_color'] = color
        verts[num_alive:]['in_size'] = new_size
        verts[num_alive:]['in_vert'] = new_pos
        self.lst.dirty = True

    def _compact(self):
        alive = self.lst.vertbuf['in_age'] < self.max_age
        self.num = num_alive = np.sum(alive)
        verts_alive = self.lst.vertbuf[alive]
        self.lst.realloc(num_alive, num_alive)
        self.lst.indexbuf[:] = np.arange(num_alive, dtype='u4') + self.lst.vertoff.start
        self.vels = self.vels[alive].copy()
        self.spins = self.spins[alive].copy()
        self.lst.vertbuf[:] = verts_alive

    def _update(self, t, dt):
        self.lst.vertbuf['in_age'] += dt
        self._compact()

        # Update
        orig_vels = self.vels
        self.vels = self.vels * self.drag ** dt + self.gravity * dt

        self.lst.vertbuf['in_vert'] += (self.vels + orig_vels) * (dt * 0.5)
        self.lst.vertbuf['in_angle'] += self.spins * dt
        self.lst.dirty = True

    def _migrate(self, vao: VAO):
        """Migrate the particles into the given VAO."""
        self.vao = vao
        num = max(self.num, 1024)
        idxs = np.arange(num, dtype='u4')
        # Allocate a large slice (set high water mark)
        self.lst = vao.alloc(num, num)
        self.lst.indexbuf[:] = idxs + self.lst.vertoff.start

        # Realloc to how much we actually want
        self.lst.realloc(self.num, self.num)

    def delete(self):
        self.layer.objects.discard(self)
        self.layer._dynamic.discard(self)
        self.lst.delete()



import unicodedata
import pkgutil
import io

import moderngl
import numpy as np
import pygame.font

from ..allocators.vertlists import VAO
from ..loaders import fonts
from ..sprites import Colorable, Transformable, TEXTURED_QUADS_PROGRAM, QUAD
from ..atlas import Atlas


FONT_LOAD_SIZE = 48
default_font = None


def get_default_font():
    """Get the default font."""
    global default_font
    if not default_font:
        ttf_data = pkgutil.get_data('wasabi2d', 'data/roboto_regular.ttf')
        default_font = pygame.font.Font(io.BytesIO(ttf_data), FONT_LOAD_SIZE)
    return default_font


class TextureVAO(VAO):
    """A VAO with an associated texture."""

    def render(self):
        self.prog['tex'].value = 0
        self.tex.use(0)
        super().render()


# Slight modification to the textured quads: we only use the texture's alpha
# channel because pygame pre-multiplies alpha in text rendering.
TEXT_PROGRAM = {
    **TEXTURED_QUADS_PROGRAM,
    'fragment_shader': '''
        #version 330

        out vec4 f_color;
        in vec2 uv;
        in vec4 color;
        uniform sampler2D tex;

        void main() {
            f_color = vec4(color.rgb, color.a * texture(tex, uv).a);
        }
    ''',
}


def text_vao(
        ctx: moderngl.Context,
        shadermgr: 'wasabi2d.layers.ShaderManager') -> VAO:
    """Build a BAO for rendering plain colored objects."""
    return TextureVAO(
        mode=moderngl.TRIANGLES,
        ctx=ctx,
        prog=shadermgr.get(**TEXT_PROGRAM),
        dtype=np.dtype([
            ('in_vert', '2f4'),
            ('in_color', '4f4'),
            ('in_uv', '2f4'),
        ])
    )


class FontAtlas(Atlas):
    """The combination of a font and the texture atlas it uses."""
    def __init__(self, ctx, font_name):
        super().__init__(ctx)
        if font_name is None:
            self.font = get_default_font()
        else:
            self.font = fonts.load(font_name, fontsize=FONT_LOAD_SIZE)

    def set_anchor(self, verts, w, h):
        """Set the anchor position to the bottom-left of the glyph."""
        verts -= (w, h, 0)

    def _load(self, name):
        """Load the image for the given name."""
        return self.font.render(name, True, (255, 255, 255))


# Text alignments as a fraction of text layout width
ALIGNMENTS = {
    'left': 0,
    'center': 0.5,
    'right': 1,
}


class Label(Colorable, Transformable):
    """A single-line text block with no additional layout/wrapping."""

    def __init__(
            self,
            text: str,
            font_atlas: FontAtlas,
            layer,
            *,
            align='left',
            fontsize=20,
            pos=(0, 0),
            color=(1, 1, 1, 1)):
        assert align in ALIGNMENTS, f"{align!r} is not a valid text alignment."
        super().__init__()
        self.lst = None
        self._verts = None
        self.align = align
        self.layer = layer
        self.fontsize = fontsize
        self.font_atlas = font_atlas
        self.pos = pos
        self.color = color
        self.text = text  # trigger layout

    def _set_dirty(self):
        self.layer._dirty.add(self)

    @property
    def align(self):
        """Get the alignment of the text relative to pos."""
        return self._align

    @align.setter
    def align(self, align):
        if align not in ALIGNMENTS:
            raise ValueError(f"{align!r} is not a valid text alignment.")
        self._align = align
        if self._verts is not None:
            self._layout()

    @property
    def text(self) -> str:
        """Get the text of this label."""
        return self._text

    @text.setter
    def text(self, text: str):
        """Set the text of this label.

        The string is currently normalised to composed form because this
        is simpler to lay out than a string with combining characters, and
        we may not be very sophisticated at this point.

        """
        text = unicodedata.normalize('NFKC', text)
        self._text = text
        self._layout()

        self._set_dirty()

    def _layout(self):
        """Generate indices, uvs and verts for the text."""
        font = self.font_atlas.font

        descent = font.get_descent()
        n_chars = len(self._text) - self._text.count('\n')
        verts = np.ones((4 * n_chars, 3), dtype='f4')
        uvs = np.zeros((4 * n_chars, 2), dtype='f4')
        indices = np.zeros(n_chars * 6, dtype='u4')

        lines = self._text.split('\n')

        # We lay out based on 48px tex so line size is scaled later
        line_height = 48 * 1.3

        curchar = 0
        for lineno, line in enumerate(lines):
            if not line:
                continue
            # (min_x, max_x, min_y, max_y, horizontal_advance_x)
            metrics = np.array(font.metrics(line), dtype='f4')
            cx = np.cumsum(metrics[:, 4]).reshape(-1, 1)
            xpos = metrics[:, 0:2] + cx

            layout_width = cx[-1] + metrics[-1, 1]
            align_offset = ALIGNMENTS[self._align] * layout_width
            yoff = lineno * line_height

            tex = None
            tex_ids = set()
            for idx, char in enumerate(line):
                tex, glyph_uvs, glyph_verts = self.font_atlas.get(char)
                tex_ids.add(tex.glo)

                # The kerning seems pretty bad on Pygame fonts...
    #            glyph_width = glyph_verts[1, 0] - glyph_verts[0, 0]
    #            metrics_width = xpos[idx, 1] - xpos[idx, 0]
    #            print(repr(char), glyph_width, metrics_width, metrics[idx, 4])

                x = xpos[idx, 0]

                quadnum = idx + curchar
                start = quadnum * 4
                glyph_slice = slice(start, start + 4)
                verts[glyph_slice] = glyph_verts + (x - align_offset, yoff - descent, 0)
                uvs[glyph_slice] = glyph_uvs
                indices[6 * quadnum:6 * quadnum + 6] = QUAD + 4 * quadnum

            curchar += len(line)

        # Scale coordinates
        resize = np.identity(3, dtype='f4')
        scale = self.fontsize / 48
        resize[0, 0] = scale
        resize[1, 1] = scale

        # TODO: handle use of multiple textures. We will be able to handle
        # this eventually by selecting texture unit within the shader, or by
        # making multiple draw calls
        assert len(tex_ids) == 1, "Label got allocated over multiple textures"
        self.tex = tex
        self._verts = verts @ resize
        self._uvs = uvs
        self._indices = indices

        # TODO: update self.lst, set dirty OR reallocate self.lst for new size
        if self.lst:
            self.lst.realloc(len(self._verts), len(indices))
            self.lst.indexbuf[:] = indices + self.lst.vertoff.start
            self.lst.vertbuf['in_uv'] = uvs
            self._update()

    def _update(self):
        xform = self._scale @ self._rot @ self._xlate

        np.matmul(
            self._verts,
            xform[:, :2],
            self.lst.vertbuf['in_vert']
        )
        self.lst.vertbuf['in_color'] = self._color
        self.lst.dirty = True

    def _migrate(self, vao: TextureVAO):
        """Migrate the fill into the given VAO."""
        # TODO: dealloc from an existing VAO
        idxs = self._indices
        self.vao = vao
        self.vao.tex = self.tex
        self.lst = vao.alloc(len(self._verts), len(idxs))
        self.lst.indexbuf[:] = idxs
        self.lst.vertbuf['in_uv'] = self._uvs
        self._update()



"""Pack sprites into texture atlases."""
from typing import Tuple, Iterable
from itertools import product
from operator import attrgetter

import numpy as np
import pygame
from pygame import Rect

from .loaders import images


class NoFit(Exception):
    """The block does not fit inside a texture."""


class BaseTex:
    """Base class for packing algorithms.

    In all cases the approach is the same:

    * consider placing the rect into each of the "free" regions.
    * score each of these and pick the best
    * update the free regions based on the selected solution

    The algorithms differ in how they score each solution and how they manage
    available free regions after a placement.

    Note that *Tex classes do not keep track of the *allocated* objects within
    them, only the free space.

    """
    def __init__(self, size: int = 512):
        self.bounds = Rect(0, 0, size, size)
        self.free = [Rect(self.bounds)]

    def _solutions(self, r: Rect) -> Iterable[Tuple[Rect, Rect]]:
        """Iterate over possible locations for r in this texture."""
        rotations = [r]
        if r.h != r.w:
            rotations.append(Rect(0, 0, r.h, r.w))

        return self._validate(product(rotations, self.free))

    def _validate(self,
                  solutions: Iterable[Tuple[Rect, Rect]]
                  ) -> Iterable[Tuple[Rect, Rect]]:
        """Check to see if solutions are valid."""
        for r, block in solutions:
            if block.w >= r.w and block.h >= r.h:
                yield r, block

    def _fitness(self, sol: Tuple[Rect, Rect]) -> float:
        """Return a value indicating the fitness of this location.

        sol is a tuple of (rect, free_region).

        Lower is better; the minimal value will be picked.
        """
        raise NotImplementedError("Subclasses must implement _fitness().")

    def _manage_free(self, rect: Rect, block: Rect):
        """Update the free lists given that rect has been placed into block."""
        raise NotImplementedError("Subclasses must implement _manage_free().")

    def place(self, r: Rect) -> Rect:
        """Place r into this texture.

        Raise NoFit if space could not be found.
        """
        if not self.free:
            raise NoFit()

        try:
            r, block = min(
                self._solutions(r),
                key=self._fitness
            )
        except ValueError as e:
            if e.args[0] != 'min() arg is an empty sequence':
                raise
            raise NoFit() from None

        r.topleft = block.topleft
        self._manage_free(Rect(r), block)
        return r


class MaxRectsTex(BaseTex):
    """Implement the MaxRects algorithm.

    MaxRects tracks all available space in a texture, potentially in very long
    list of overlapping free regions. Adding a rect partitions any intersecting
    free regions so that they do not intersect - potentially creating up to
    4 new free regions for each intersecting free region.

    Packing is very good but can be very slow, particularly when the size of
    rects requested is a small fraction of the available space; this tends to
    create very large numbers of free regions.

    This implementation takes around 40ms to pack 500 randomly sized sprites,
    but as much as 4100ms to pack 500 font glyphs with a low size distribution.

    To control this growth a parameter threshold limits the size of areas we
    keep track of. An area whose smallest dimension is less than threshold is
    not tracked - possibly other larger overlapping regions contain some of the
    space. Threshold values between 16 and 32 give significant speedups without
    badly affecting packing efficiency.

    A reasonable description of MaxRects is here:

    https://eatplayhate.me/2013/09/17/adventures-in-engine-construction-rectangle-packing/

    """

    def __init__(self, size: int = 512, threshold: int = 0):
        self.largest_dim = size
        self.threshold = threshold
        super().__init__(size)

    def _validate(self, solutions):
        for r, block in solutions:
            if block.w < r.w:
                # We keep list sorted so we can break here
                break
            elif block.h < r.h:
                continue

            yield r, block

    def _fitness(self, sol):
        """Return a value indicating the fitness of this localstion.

        Lower is better; the minimal value will be picked.
        """
        r, block = sol
        # Best short side fit
        if r.w < r.h:
            return block.w - r.w
        else:
            return block.h - r.h

    def _manage_free(self, r, _):
        new_free = []  # new rects already checked against threshold
        new_rects = []  # new rects to be checked
        for block in self.free:
            if not block.colliderect(r):
                new_free.append(block)
                continue

            new_rects.extend([
                # Above r
                Rect(block.left, block.top, block.w, r.top - block.top),
                # Below r
                Rect(block.left, r.bottom, block.w, block.bottom - r.bottom),
                # Left of r
                Rect(block.left, block.top, r.left - block.left, block.h),
                # Right of r
                Rect(r.right, block.top, block.right - r.right, block.h),
            ])

        new_free += [
            r for r in new_rects
            if min(r.size) > self.threshold
        ]

        # This is O(n²) but I couldn't improve it by a sort-and sweep approach
        # due to high constant factors
        self.free = [
            r for r in new_free
            if not any(b.contains(r) for b in new_free if b is not r)
        ]
        self.free.sort(key=attrgetter('w'), reverse=True)


class ShelvesTex(BaseTex):
    """Use the shelf algorithm to pack a texture.

    Shelves split the vertical space into "shelf" regions and split the
    horizontal space among objects on a shelf.

    While it wastes more space than MaxRects, this algorithm is much faster.

    This implementation takes around ~10ms to pack 500 font-glyph-size objects
    and a similar amount of time to pack 500 more randomly sized sprites. In
    the case of packing font glyphs, where the size distribution is small, the
    packs are also relatively tight.
    """

    def __init__(self, size=512):
        self.bounds = Rect(0, 0, size, size)
        self.contents = []
        self.free = [Rect(self.bounds)]

    def _validate(self, solutions):
        """Validate the solutions.

        We override this to add a termination condition.
        """
        for r, block in solutions:
            if block.h < r.h:
                # We keep list sorted so we can break here
                break
            elif block.w < r.w:
                continue

            yield r, block

    def _fitness(self, sol):
        """Return a value indicating the fitness of this localstion.

        Lower is better; the minimal value will be picked.
        """
        r, block = sol
        if block.left == 0:
            # Weight creating a new shelf
            return r.h * 0.2
        return block.h - r.h

    def _manage_free(self, r, block):
        if block.left == 0:
            # Allocating a new shelf
            top = block.top
            block.height -= r.height
            block.top = r.bottom
            if block.w > r.w:
                self.free.append(Rect(r.right, top, block.w - r.w, r.h))
            if block.height <= 0:
                self.free.remove(block)
        else:
            block.width -= r.width
            block.left = r.right
            if block.width <= 0:
                self.free.remove(block)

        self.free.sort(key=attrgetter('h'), reverse=True)


class Packer:
    @classmethod
    def new_maxrects(cls, size=512, threshold=0):
        return cls(
            tex_factory=lambda: MaxRectsTex(
                size=size,
                threshold=threshold
            )
        )

    @classmethod
    def new_shelves(cls, size=512):
        return cls(tex_factory=lambda: ShelvesTex(size))

    def __init__(self, tex_factory):
        self.texs = []
        self.new_tex = tex_factory

    def add(self, r):
        i = -1
        for i, t in enumerate(self.texs):
            try:
                result = t.place(r)
            except NoFit:
                continue
            else:
                return i, result
        new = self.new_tex()
        result = new.place(r)
        self.texs.append(new)
        return i + 1, result


class Atlas:
    def __init__(self, ctx, *, texsize=512, padding=2):
        self.ctx = ctx
        self.padding = padding
        self.texsize = texsize

        self.packer = Packer.new_shelves(size=texsize)

        self.surfs_texs = []
        self._dirty = set()
        self.tex_for_name = {}

    def _load(self, name):
        """Load the image for the given name."""
        return images.load(name)

    def mksurftex(self, size):
        """Make a new surface and corresponding texture of the given size."""
        surf = pygame.Surface(size, pygame.SRCALPHA, depth=32)
        tex = self.ctx.texture(size, 4)
        self.surfs_texs.append((surf, tex))
        return surf, tex

    def npot_tex(self, sprite_name, img):
        """Get a non-power-of-two texture for this image."""
        w, h = size = img.get_size()
        tex = self.ctx.texture(size, 4)
        tex.write(pygame.image.tostring(img, "RGBA", 1))
        tex.build_mipmaps(max_level=2)
        texcoords = np.array([
            (0, 1),
            (1, 1),
            (1, 0),
            (0, 0),
        ], dtype='f4')
        verts = np.array([
            (0, 0, 1),
            (w, 0, 1),
            (w, h, 1),
            (0, h, 1),
        ], dtype='f4')
        self.set_anchor(verts, w, h)
        res = self.tex_for_name[sprite_name] = (tex, texcoords, verts)
        return res

    def get(self, sprite_name):
        if sprite_name in self.tex_for_name:
            return self.tex_for_name[sprite_name]

        img = self._load(sprite_name)

        pad = self.padding * 2

        orig = img.get_rect()
        r = Rect(orig)
        r.w += pad
        r.h += pad
        try:
            bin, packed = self.packer.add(r)
        except NoFit:
            # Does not fit in packer bounds
            return self.npot_tex(sprite_name, img)

        p = Rect(packed)
        p.left += self.padding
        p.top += self.padding
        p.w -= pad
        p.h -= pad

        try:
            surf, tex = self.surfs_texs[bin]
        except IndexError:
            size = (self.texsize, self.texsize)
            surf, tex = self.mksurftex(size)
        self._dirty.add(bin)

        rotated = False
        if orig.w != p.w:
            img = pygame.transform.rotate(img, -90)
            rotated = True

        x, y = p.topleft
        w, h = p.size

        surf.blit(img, p)

        l = p.left / self.texsize
        t = 1.0 - p.top / self.texsize
        r = p.right / self.texsize
        b = 1.0 - p.bottom / self.texsize

        if rotated:
            texcoords = np.array([
                (r, t),
                (r, b),
                (l, b),
                (l, t),
            ], dtype='f4')
        else:
            texcoords = np.array([
                (l, t),
                (r, t),
                (r, b),
                (l, b),
            ], dtype='f4')
        verts = np.array([
            (0, 0, 1),
            (orig.w, 0, 1),
            (orig.w, orig.h, 1),
            (0, orig.h, 1),
        ], dtype='f4')
        self.set_anchor(verts, orig.w, orig.h)
        res = self.tex_for_name[sprite_name] = (tex, texcoords, verts)
        return res

    def set_anchor(self, verts, w, h):
        """Set the anchor position to the center of the sprite."""
        verts -= (w / 2, h / 2, 0)

    def _update(self):
        """Copy updated surfaces to the GL texture objects."""
        for bin in self._dirty:
            surf, tex = self.surfs_texs[bin]
            tex.write(pygame.image.tostring(surf, "RGBA", 1))
            tex.build_mipmaps(max_level=2)
        self._dirty.clear()

    def dump(self):
        """Save screenshots of all the textures."""
        for i, (surf, _) in enumerate(self.surfs_texs):
            fname = f'atlas{i}.png'
            pygame.image.save(surf, fname)



import random
from wasabi2d import Scene, run, event, Storage, Vector2
from wasabi2d.actor import Actor


scene = Scene(
    width=400,
    height=708,
    title='Flappy Bird',
)
storage = Storage()
storage.setdefault('highscore', 0)


# These constants control the difficulty of the game
GAP = 130
GRAVITY = 0.3
FLAP_STRENGTH = 6.5
SPEED = 3


score_label = scene.layers[5].add_label(
    "0",
    color='white',
    pos=(scene.width / 2, 80),
    align='center',
    fontsize=70,
)


class Score:
    _score = 0

    @property
    def score(self):
        return self._score

    @score.setter
    def score(self, s):
        self._score = s
        score_label.text = str(s)


score = Score()


scene.layers[-2].add_sprite(
    'background',
    pos=(scene.width / 2, scene.height / 2)
)
bird = Actor(
    scene.layers[0].add_sprite('bird1'),
    pos=(75, 200)
)
bird.dead = False
bird.vy = 0


class Pipes:
    def __init__(self):
        self.top = scene.layers[-1].add_sprite('top', anchor=('left', 'bottom'))
        self.bottom = scene.layers[-1].add_sprite('bottom', anchor=('left', 'top'))
        self.x = scene.width
        self.w = self.top.width
        self.gap = 0
        self.pipe_h = self.top.height

    @property
    def x(self):
        return self._x

    @x.setter
    def x(self, x):
        self._x = x
        self.top.x = x
        self.bottom.x = x

    def set_gap(self, y):
        self.gap = y
        self.top.y = y - GAP // 2 - self.pipe_h / 2
        self.bottom.y = y + GAP // 2 + self.pipe_h / 2


pipes = Pipes()
highscore_label = scene.layers[5].add_label(
    f"Best: {storage['highscore']}",
    color=(200, 170, 0),
    align='center',
    pos=(scene.width / 2, scene.height - 10),
    fontsize=30,
)


def reset_pipes():
    pipes.set_gap(random.randint(200, scene.height - 200))
    pipes.x = scene.width + pipes.w


reset_pipes()  # Set initial pipe positions.


def update_pipes():
    pipes.x -= SPEED

    if pipes.x < -0.5 * pipes.w:
        reset_pipes()
        if not bird.dead:
            score.score += 1
            if score.score > storage['highscore']:
                storage['highscore'] = score.score
                highscore_label.text = f"Best: {score.score}"


def update_bird():
    uy = bird.vy
    bird.vy += GRAVITY
    bird.y += (uy + bird.vy) / 2

    if not bird.dead:
        if bird.vy < -3:
            bird.image = 'bird2'
        else:
            bird.image = 'bird1'

    px = pipes.x

    pipes_left = px - pipes.w / 2
    pipes_right = px + pipes.w / 2
    if pipes_left < bird.right and bird.left < pipes_right \
            and not pipes.gap - GAP // 2 < bird.y < pipes.gap + GAP // 2:
        bird.dead = True
        bird.image = 'birddead'

    if not 0 < bird.y < 720:
        bird.y = 200
        bird.dead = False
        score.score = 0
        bird.vy = 0
        reset_pipes()


@event
def update():
    update_pipes()
    update_bird()


@event
def on_key_down():
    if not bird.dead:
        bird.vy = -FLAP_STRENGTH


run()



from typing import Tuple, List
from dataclasses import dataclass

import moderngl
import numpy as np

from .base import PostprocessPass


def gaussian(x, mu, sig):
    """Calculate a gaussian function."""
    return np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))


THRESHOLD_PROG = """ \
#version 330 core

in vec2 uv;
out vec4 f_color;

uniform sampler2D image;

void main()
{
    vec4 in_col = texture(image, uv);
    float lum = dot(vec3(0.3, 0.6, 0.1), in_col.rgb) * in_col.a;
    f_color = in_col * pow(lum, 2);
}

"""


COPY_PROG = """ \
#version 330 core

in vec2 uv;
out vec4 f_color;
uniform sampler2D image;

void main()
{
    f_color = texture(image, uv);
}
"""

# Shader code adapted from https://learnopengl.com/Advanced-Lighting/Bloom
# First pass, blur vertically
BLUR_PROG = """ \
#version 330 core

in vec2 uv;
out vec4 f_color;

uniform sampler2D image;
uniform sampler2D gauss_tex;
uniform float radius;
uniform vec2 blur_direction;


float gauss(float off) {
    return texture(gauss_tex, vec2(abs(off / radius), 0)).r;
}


void main()
{
    vec2 tex_offset = 1.0 / textureSize(image, 0); // gets size of single texel
    vec3 result = texture(image, uv).rgb; // current fragment's contribution

    vec2 lookup_stride = tex_offset * blur_direction;
    float weight_sum = 1.0;
    float weight;
    int irad = int(radius);
    for(int i = 1; i < irad; ++i)
    {
        weight = gauss(i);
        weight_sum += i * 2;
        result += texture(image, uv + lookup_stride * i).rgb * weight;
        result += texture(image, uv - lookup_stride * i).rgb * weight;
    }
    f_color = vec4(result / radius * 2, 1);
}

"""


@dataclass
class Bloom:
    """A light bloom effect."""
    ctx: moderngl.Context
    shadermgr: 'wasabi2d.layers.ShaderManager'
    threshold: float = 1.0
    radius: float = 10.0

    camera: 'wasabi2d.scene.Camera' = None
    _blur: PostprocessPass = None
    _fb1: moderngl.Framebuffer = None
    _fb2: moderngl.Framebuffer = None

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera
        self._fb1, = camera._get_temporary_fbs(1, 'f2')
        self._thresholded = camera._make_fb('f2', div_x=4, div_y=4)
        gauss = gaussian(np.arange(256), 0, 90).astype('f4')
        self._gauss_tex = self.ctx.texture((256, 1), 1, data=gauss, dtype='f4')
        self._fb2 = camera._make_fb('f2', div_x=4)
        self._threshold_pass = PostprocessPass(
            self.ctx,
            self.shadermgr,
            THRESHOLD_PROG,
        )
        self._blur = PostprocessPass(
            self.ctx,
            self.shadermgr,
            BLUR_PROG
        )
        self._copy = self._mkpass(COPY_PROG)

    def _mkpass(self, shader):
        return PostprocessPass(self.ctx, self.shadermgr, shader)

    def enter(self, t, dt):
        self._fb1.use()
        self._fb1.clear()

    def exit(self, t, dt):
        self._thresholded.use()
        self._thresholded.clear()
        self._threshold_pass.render(image=self._fb1)

        self._fb2.use()
        self._fb2.clear()
        self._blur.render(
            image=self._thresholded,
            blur_direction=(0, 1),
            radius=self.radius,
            gauss_tex=self._gauss_tex,
        )
        self.ctx.screen.use()

        self._copy.render(image=self._fb1)
        self.ctx.blend_func = moderngl.SRC_ALPHA, moderngl.ONE
        self._blur.render(
            image=self._fb2,
            blur_direction=(1, 0),
            radius=self.radius,
            gauss_tex=self._gauss_tex,
        )
        self.ctx.blend_func = moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA



"""An object that connects a graphics primitive with a collision shape."""
from .rect import Rect


ANCHORS = {
    'x': {
        'left': 0.0,
        'center': 0.5,
        'middle': 0.5,
        'right': 1.0,
    },
    'y': {
        'top': 0.0,
        'center': 0.5,
        'middle': 0.5,
        'bottom': 1.0,
    }
}


def calculate_anchor(value, dim, total):
    if isinstance(value, str):
        try:
            return total * ANCHORS[dim][value]
        except KeyError:
            raise ValueError(
                '%r is not a valid %s-anchor name' % (value, dim)
            )
    return float(value)


# These are methods (of the same name) on pygame.Rect
SYMBOLIC_POSITIONS = set((
    "topleft", "bottomleft", "topright", "bottomright",
    "midtop", "midleft", "midbottom", "midright",
    "center",
))


class Actor:
    """Actors are associated with a graphics object and a rectangle.

    We delegate attribute access to the appropriate one, and keep the two
    in sync, transferring position updates to the primitive.

    """

    RECT_ATTRS = set(Rect.VALID_ATTRIBUTES)
    PRIM_ATTRS = {'pos', 'image', 'angle'}

    def __init__(self, prim, pos=None, anchor=None, **kwargs):
        self.prim = prim  # The primitive we are wrapping

        self.__dict__["bounds"] = self.prim.bounds
        # Initialise it at (0, 0) for size (0, 0).
        # We'll move it to the right place and resize it later
        self._init_position(pos, anchor, **kwargs)
        self.prim.pos = self.bounds.center

    def __getattr__(self, name):
        """Delegate attribute access to the ZRect."""
        if name in self.PRIM_ATTRS:
            return getattr(self.prim, name)
        try:
            return getattr(self.bounds, name)
        except AttributeError:
            pass

        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        """Set an attribute.

        If the attribute is one that we know the rect class has, set it and
        then update the position of our primitive.
        """
        if name in self.RECT_ATTRS:
            setattr(self.bounds, name, value)
            self.prim.pos = self.bounds.center
        elif name in self.PRIM_ATTRS:
            setattr(self.prim, name, value)
            if name == 'image':
                self.bounds.size = self.prim.width, self.prim.height
            elif name == 'pos':
                self.bounds.center = self.prim.pos
        else:
            object.__setattr__(self, name, value)

    def _init_position(self, pos, anchor, **kwargs):
        if anchor is None:
            anchor = ("center", "center")
        self.anchor = anchor

        symbolic_pos_args = {
            k: kwargs[k] for k in kwargs if k in SYMBOLIC_POSITIONS}

        if not pos and not symbolic_pos_args:
            # No positional information given, use sensible top-left default
            self.topleft = (0, 0)
        elif pos and symbolic_pos_args:
            raise TypeError(
                "'pos' argument cannot be mixed with 'topleft', "
                "'topright' etc. argument."
            )
        elif pos:
            self.pos = pos
        else:
            self._set_symbolic_pos(symbolic_pos_args)

    def _set_symbolic_pos(self, symbolic_pos_dict):
        if len(symbolic_pos_dict) != 0:
            if len(symbolic_pos_dict) == 0:
                raise TypeError(
                    "No position-setting keyword arguments ('topleft', "
                    "'topright' etc) found."
                )
            if len(symbolic_pos_dict) > 1:
                raise TypeError(
                    "Only one 'topleft', 'topright' etc. argument is allowed."
                )

        setter_name, position = symbolic_pos_dict.popitem()
        setattr(self, setter_name, position)

    def delete(self):
        self.prim.delete()
        self.prim = None



"""wasabi2d, a 2D game engine using moderngl, pygame and numpy."""
import os

# Don't show Pygame's annoying message because while I might use PyGame,
# I don't appreciate libraries I use communicating with my users.
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'

from pygame.math import Vector2
from .game import EventMapper
from .constants import keys, mouse
from .loaders import sounds
from . import music
from . import clock
from .scene import Scene
from .animation import animate
from .storage import Storage

event = EventMapper()
del EventMapper

run = event.run

__all__ = [
    'Vector2',
    'event',
    'run',
    'keys', 'mouse',
    'sounds', 'music', 'clock', 'animate',
    'Scene', 'Storage',
]



from unittest.mock import ANY
import numpy as np

from ..color import convert_color

from ..vendor.earcut import earcut
from .base import AbstractShape

from ..rect import ZRect


class Polygon(AbstractShape):
    """An arbitrary polygon."""

    def __init__(
            self,
            layer,
            vertices,
            *,
            pos=(0, 0),
            color=(1, 1, 1, 1),
            stroke_width=1.0):

        verts = np.array(vertices, dtype='f4')
        if verts.shape != (ANY, 2):
            raise ValueError(
                f"Unsupported vertex shape {verts.shape}, expected (*, 2)"
            )

        super().__init__()
        self.layer = layer
        self.pos = pos

        self.orig_verts = np.ones((len(verts), 3), dtype='f4')
        self.orig_verts[:, :2] = verts
        self._stroke_width = stroke_width

        self._color = convert_color(color)
        self._set_dirty()

    def _stroke_indices(self):
        """Indexes for drawing the stroke as a LINE_STRIP."""
        verts = len(self.orig_verts)
        idxs = np.arange(verts + 1, dtype='i4')
        idxs[-1] = 0
        return idxs[[-1, *range(verts), 0, 1, 2]]

    def _fill_indices(self):
        """Indexes for drawing the fill as TRIANGLES."""
        verts = self.orig_verts[:, :2].reshape((-1))
        idxs = earcut(verts)
        return np.array(idxs, dtype='i4')


class Rect(Polygon):
    """A rectangle."""
    VERTS = np.array([
        (-0.5, -0.5),
        (-0.5, 0.5),
        (0.5, 0.5),
        (0.5, -0.5),
    ], dtype='f4')

    __slots__ = (
        'width', 'height',
    )

    def __init__(
            self,
            layer,
            width,
            height,
            *,
            pos=(0, 0),
            color=(1, 1, 1, 1)):

        self.width = width
        self.height = height

        super().__init__(
            layer,
            self._vertices(),
            pos=pos,
            color=color
        )

    def _vertices(self):
        return np.copy(self.VERTS) * (self.width, self.height)


class PolyLine(Polygon):
    """An unclosed line sequence."""

    def _stroke_indices(self):
        """Indexes for drawing the line as a LINE_STRIP."""
        verts = len(self.orig_verts)
        idxs = np.arange(verts)
        last = verts - 1
        return [0, *idxs, last, last]

    def _fill_indices(self):
        raise NotImplementedError(
            "Lines may not be filled."""
        )



import json
import os
import platform
from hashlib import sha1


__all__ = ['StorageCorruptionException', 'JSONEncodingException', 'Storage']


class StorageCorruptionException(Exception):
    """The data in the storage is corrupted."""


class JSONEncodingException(Exception):
    """The data in the storage is corrupted."""


def _get_platform_wasabi2d_path():
    r"""Get the storage directory for wasabi2d save data.

    Under Windows, return %APPDATA%\wasabi2d. Under Linux/MacOS, return
    ~/.config/wasabi2d/saves.

    """
    if platform.system() == 'Windows':
        try:
            appdata = os.environ['APPDATA']
        except KeyError:
            raise KeyError(
                "Couldn't find the AppData directory for Pygame Zero save "
                "data. Please set the %APPDATA% environment variable."
            )
        return os.path.join(appdata, 'wasabi2d')
    return os.path.expanduser(os.path.join('~', '.config/wasabi2d/saves'))


class Storage(dict):
    """Behaves like a dictionary that serialises itself to disk.

    The name of the file will be based on the basename of the script plus
    a hash of the script's path, ensuring that each script on the filesystem
    has a unique save file.

    """
    STORAGE_DIR = _get_platform_wasabi2d_path()

    # Keep a reference to all defined storages
    storages = []

    def __init__(self, filename=None):
        super().__init__()
        self.loaded = False
        self._save_file = filename
        self.storages.append(self)

    @classmethod
    def save_all(cls):
        """Save all instances of Storage."""
        for storage in cls.storages:
            storage.save()

    @classmethod
    def _ensure_save_path(cls):
        """Ensure that the directory for all save game data exists."""
        storage_path = cls._get_platform_wasabi2d_path()
        try:
            os.makedirs(storage_path)
        except IsADirectoryError:
            pass

    def _set_filename_from_path(self, file_path):
        """Set the path to save to from the given filename.

        We include the basename of the game script, to help with identifying
        the save data, as well as the sha1 hash of the full path of the file on
        disk, to ensure that every separate game has its own storage.

        """
        if not os.path.isabs(file_path):
            file_path = os.path.abspath(file_path)

        fn_hash = sha1(file_path.encode('utf-8')).hexdigest()
        base, _ = os.path.splitext(os.path.basename(file_path))
        self._save_file = '{}-{}.json'.format(base, fn_hash)
        self.load()

    @property
    def path(self):
        """Get the path for data data."""
        if self._save_file is None:
            raise ValueError(
                "Cannot save/load storage as no filename is set."
            )
        return os.path.join(self.STORAGE_DIR, self._save_file)

    def load(self):
        """Load data into the storage from disk.

        This replaces all previous contents of the storage. If there is no save
        file found then the storage will be empty.

        """
        self.clear()

        try:
            with open(self.path, 'r') as f:
                data = json.load(f)
        except FileNotFoundError:
            # If no file exists, it's fine
            pass
        except json.JSONDecodeError:
            raise StorageCorruptionException(
                "Storage is corrupted. Couldn't load the data."
            )
        else:
            self.update(data)
            self.loaded = True

    def save(self):
        """Save data to disk."""
        if not self and not self.loaded:
            return
        try:
            data = json.dumps(self)
        except TypeError:
            msgs = [
                "{} - type {}".format(*item)
                for item in self._get_json_error_keys(self)
            ]
            if not msgs:
                # Didn't find an explanation, so let original error propagate
                raise
            raise JSONEncodingException(
                "The following storage entries cannot be stored, because "
                "they are not JSON serializable types:\n{}\n\n"
                "Only the following types of objects are JSON-serialisable: "
                "dict, list/tuple, str, float/int, bool, and None".format(
                    "\n".join(msgs)
                )
            )
        else:
            path = self.path
            with open(path, 'w+') as f:
                f.write(data)
            print("Saved storage to", path)

    # Constants for use with isinstance in _get_json_error_keys()
    JSON_PRIMITIVES = (float, int, str, bool, type(None))
    JSON_SEQ = (list, tuple)
    JSON_MAPPING = dict

    @classmethod
    def _get_json_error_keys(cls, obj, json_path='storage'):
        """Identify the paths with the storage which failed to be JSON encoded.

        Return an iterable of message strings.

        """
        if isinstance(obj, dict):
            # TODO: also check/report that k is a string, as only strings are
            # valid as object keys in JSON.
            for k, v in obj.items():
                if isinstance(v, cls.JSON_PRIMITIVES):
                    continue
                subpath = '{}[{!r}]'.format(json_path, k)
                yield from cls._get_json_error_keys(v, subpath)
        elif isinstance(obj, (list, tuple)):
            for i, value in enumerate(obj):
                if isinstance(value, cls.JSON_PRIMITIVES):
                    continue
                subpath = '{}[{}]'.format(json_path, i)
                yield from cls._get_json_error_keys(value, subpath)
        elif isinstance(obj, cls.JSON_PRIMITIVES):
            return
        else:
            # TODO: Could have a custom JSONEncoder in the future, in which
            # case we would have to actually run json.dumps on it to get the
            # result and possibly even drill down on the parameters we got
            # back, if that object happens to have a list as an attribute, for
            # example.
            t = type(obj)
            if t.__module__ != 'builtins':
                typename = '{t.__module__}.{t.__qualname__}'.format(t=t)
            else:
                typename = t.__qualname__
            yield json_path, typename



"""Example of the light bloom effect."""
from wasabi2d import run, Scene, event, keys


scene = Scene()

logo = scene.layers[0].add_sprite(
    'wasabi2d',
    pos=(scene.width / 2, scene.height / 2),
)
logo.color = (1.3, 1.3, 1.3, 1)
scene.layers[0].set_effect(
    'bloom',
    radius=20
)


@event
def on_key_down(key, mod):
    if key == keys.F12:
        scene.screenshot()


run()



"""Color conversion functions."""
from typing import Union, Tuple
import numpy as np
from pygame import Color


def convert_color(c: Union[str, tuple]) -> Tuple[float, float, float, float]:
    """Convert a color to an RGBA tuple."""
    if isinstance(c, str):
        col = Color(c)
        return np.array(memoryview(col), dtype='u1').astype('f4') / 255.0
    else:
        assert 3 <= len(c) <= 4, "Invalid color length"
        return np.array(c + (1,) * (4 - len(c)), dtype='f4')



import typing
from typing import Tuple
from dataclasses import dataclass
from collections import OrderedDict

import moderngl
import numpy as np

from .abstract import AbstractAllocator, NoCapacity


TYPE_MAP = {
    'uint8': 'f1',
    'uint16': 'u2',
    'uint32': 'u4',
    'uint64': 'u8',
    'int8': 'i1',
    'int16': 'i2',
    'int32': 'i4',
    'int64': 'i8',
    'float16': 'f2',
    'float32': 'f4',
    'float64': 'f8',
}


def dtype_to_moderngl(dtype: np.dtype) -> tuple:
    """Convert a numpy dtype object to a ModernGL buffer type."""
    names = dtype.names
    assert names is not None, "Only structured numpy dtypes are allowed."
    fields = dtype.fields
    out = []
    byte_pos = 0  # position of next field
    for n in names:
        dtype, offset, *_ = fields[n]

        out.extend('x' * (offset - byte_pos))
        byte_pos = offset + dtype.itemsize

        type_name = TYPE_MAP[dtype.base.name]
        assert len(dtype.shape) <= 1, \
            "Multi-dimensional dtypes are not supported."
        if dtype.shape == ():
            out.append(type_name)
        else:
            out.append(f'{dtype.shape[0]}{type_name}')
    return (' '.join(out), *names)


@dataclass(eq=False)
class VAOList:
    """A list allocated within a VAO."""
    buf: 'VAO'
    command: int

    # View of the vertex buffer, and slice
    vertbuf: np.ndarray
    vertoff: slice

    # View of the index buffer, and slice
    indexbuf: np.ndarray
    indexoff: slice

    # Number of verts and indexes in the allocation
    _hwm_verts: int
    _hwm_indexes: int

    #: True if needs syncing to the GL
    dirty: bool = False

    @property
    def num_indexes(self):
        """Get the number of indices to draw."""
        pos = self.buf.allocs.index(self)
        return self.buf.indirect[pos, 0]

    @num_indexes.setter
    def num_indexes(self, n):
        """Set the number of indices to draw.

        If a larger segment is allocated, this can be used to very cheaply
        resize it.

        It is not yet supported to resize the index allocation at this point.
        """
        size = len(self.indexbuf)
        if n > size:
            raise ValueError(f"Only allocated {size} indices.")

        # TODO: linear cost
        pos = self.buf.allocs.index(self)
        self.buf.indirect[pos, 0] = n

    def realloc(self, num_verts=None, num_indexes=None):
        """Reallocate the list to a new size. Invalidate the data."""
        if num_verts is None:
            num_verts = len(self.vertbuf)
        if num_indexes is None:
            num_indexes = len(self.indexbuf)

        if num_verts == len(self.vertbuf) \
                and num_indexes == len(self.indexbuf):
            return
        self.buf.realloc(self, num_verts, num_indexes)

    def free(self):
        if self.buf:
            self.buf.free(self)

    __del__ = free


class IndirectBuffer:
    """Abstraction over managing indirect allocations.

    Indirect draw commands always have the same 5 u4 parameters:

    (count, instanceCount, firstIndex, baseVertex, baseInstance)

    """
    def __init__(self, ctx, capacity=50):
        self.ctx = ctx
        self.capacity = capacity
        self.next_key = 0
        self.allocations = OrderedDict()
        self.buffer = None

    def _initialise(self):
        self.indirect = np.zeros((self.capacity, 5), dtype='u4')
        for aidx, lst in enumerate(self.allocs):
            num_indexes = lst.indexoff.stop - lst.indexoff.start
            ixs_start = lst.indexoff.start
            vs_start = lst.vertoff.start
            self.indirect[aidx] = (num_indexes, 1, ixs_start, vs_start, 0)
        self.buffer = self.ctx.buffer(self.indirect, dynamic=True)
        self.dirty = False

    def get_buffer(self):
        """Get the buffer object.

        Create a new buffer object if dirty.

        """
        if not self.buffer:
            self.buffer = self.ctx.buffer(
                np.array(list(self.allocations.values()), dtype='u4'),
            )
        return self.buffer

    def render_direct(self, vao, mode):
        cmds = self.allocations.values()
        for vs, insts, base_idx, base_v, base_inst in cmds:
            vao.render(mode, vs, first=base_idx, instances=1)

    def append(self, vs, insts, base_idx, base_v, base_inst) -> int:
        """Append an indirect draw command.

        Return an opaque key that can be used to update or delete the command.
        """
        assert base_v == 0
        key = self.next_key
        self.next_key += 1

        self.allocations[key] = np.array(
            [vs, insts, base_idx, base_v, base_inst],
            dtype='u4'
        )
        self.buffer = None
        return key

    def __delitem__(self, key):
        del self.allocations[key]
        self.buffer = None

    def __setitem__(self, key, vals):
        assert len(vals) == 5, "Invalid indirect draw command"
        self.allocations[key][:] = vals
        self.buffer = None


class MemoryBackedBuffer:
    """Maintain a GL buffer plus a numpy arrays for storage."""

    def __init__(self, ctx, capacity, dtype):
        self.ctx = ctx
        self.allocator = AbstractAllocator(capacity)
        self.dtype = dtype
        self.array = np.empty(capacity, dtype=self.dtype)
        self.buffer = None

    def allocate(self, num: int) -> Tuple[slice, np.ndarray]:
        """Allocate a slice of the array.

        Return a slice and view.

        """
        try:
            allocated = self.allocator.alloc(num)
        except NoCapacity as e:
            new_size = e.recommended
            new_array = np.empty(new_size, dtype=self.dtype)
            new_array[:len(self.array)] = self.array
            self.array = new_array
            self.allocator.grow(e.recommended)
            self.buffer = None
            allocated = self.allocator.alloc(num)
        return allocated, self.array[allocated]

    def get_buffer(self, dirty=False) -> moderngl.Buffer:
        """Get the buffer."""
        if not self.buffer:
            self.buffer = self.ctx.buffer(self.array, dynamic=True)
        elif dirty:
            self.buffer.write(self.array)
        return self.buffer

    def realloc(self, offset: slice, size: int) -> Tuple[slice, np.ndarray]:
        """Resize the allocation at offset. Return the new slice and view."""
        # TODO: use self.allocator.realloc here
        self.allocator.free(offset)
        newoff, new_view = self.allocate(size)
        copy_size = min(size, offset.stop - offset.start)
        new_view[:copy_size] = self.array[offset][:copy_size]
        return newoff, new_view

    def free(self, offset: typing.Union[int, slice]):
        """Free the allocation."""
        self.allocator.free(offset)


class VAO:
    """Manage vertex lists within a VAO."""

    def __init__(
            self,
            mode: int,
            ctx: moderngl.Context,
            prog: moderngl.Program,
            dtype: np.dtype,
            capacity: int = 4096,
            index_capacity: int = 8192):
        self.mode = mode
        self.ctx = ctx
        self.prog = prog
        self.dtype = dtype_to_moderngl(dtype)
        self.verts = MemoryBackedBuffer(ctx, capacity, dtype)
        self.indexes = MemoryBackedBuffer(ctx, index_capacity, 'i4')
        self.indirect = IndirectBuffer(ctx)
        self.allocs: typing.List[VAOList] = []
        self.vao = None

    def _initialise(self):
        """Create OpenGL objects."""
        self.vbo = self.verts.get_buffer()
        self.ibo = self.indexes.get_buffer()
        self.vao = self.ctx.vertex_array(
            self.prog,
            [
                (self.vbo, *self.dtype),
            ],
            self.ibo
        )

    def alloc(self, num_verts: int, num_indexes: int) -> VAOList:
        """Allocate a list from within this buffer."""
        vs, vertbuf = self.verts.allocate(num_verts)
        ixs, indexbuf = self.indexes.allocate(num_indexes)

        cmd = self.indirect.append(num_indexes, 1, ixs.start, 0, 0)
        lst = VAOList(
            buf=self,
            command=cmd,
            vertbuf=vertbuf,
            vertoff=vs,
            indexbuf=indexbuf,
            indexoff=ixs,
            _hwm_verts=num_verts,
            _hwm_indexes=num_indexes,
            dirty=True
        )
        self.allocs.append(lst)
        return lst

    def realloc(self, lst: VAOList, num_verts: int, num_indexes: int):
        """Reallocate a list, typically to grow the size of it.

        Data will be copied to the new list and it will retain its draw
        order.

        We maintain high water marks on the allocation for a list and will
        shrink an existing numpy slice without going back to the allocator.
        When we do go back, we ask for extra.
        """
        need_verts = num_verts != len(lst.vertbuf)
        need_idxs = num_indexes != len(lst.indexbuf)

        if need_verts:
            lst.vertoff, lst.vertbuf = self.verts.realloc(
                lst.vertoff,
                num_verts,
            )
        if need_idxs:
            lst.indexoff, lst.indexbuf = self.indexes.realloc(
                lst.indexoff,
                num_indexes,
            )

        lst.dirty = True
        self.indirect[lst.command] = (
            num_indexes,
            1,
            lst.indexoff.start,
            lst.vertoff.start,
            0
        )

    def free(self, lst):
        """Remove a list from the array."""
        self.allocs.remove(lst)
        del self.indirect[lst.command]

        # Free space in allocators
        self.verts.free(lst.vertoff)
        self.indexes.free(lst.indexoff)

        lst.buf = None
        lst.vertbuf = None
        lst.indexbuf = None

    def get_vao(self):
        # TODO: use the dirty list to more accurately indicate which parts of
        # a buffer need updating.
        dirty = False
        for a in self.allocs:
            if a.dirty:
                dirty = True
                a.dirty = False

        vbo = self.verts.get_buffer(dirty)
        ibo = self.indexes.get_buffer(dirty)

        # TODO: only recreate the VAO if buffers have changed
        vao = self.ctx.vertex_array(
            self.prog,
            [
                (vbo, *self.dtype),
            ],
            ibo
        )
        return vao

    def render(self):
        """Render all lists."""
        if not self.allocs:
            return
        vao = self.get_vao()
        if self.ctx.version_code >= 400:
            indirect = self.indirect.get_buffer()
            vao.render_indirect(
                indirect,
                mode=self.mode,
            )
        else:
            self.indirect.render_direct(vao, self.mode)



import numpy as np

from .primitives.base import Colorable, Transformable, Bounds


TEXTURED_QUADS_PROGRAM = dict(
    vertex_shader='''
        #version 330

        uniform mat4 proj;

        in vec3 in_vert;
        in vec4 in_color;
        in vec2 in_uv;
        out vec2 uv;
        out vec4 color;

        void main() {
            gl_Position = proj * vec4(in_vert.xy, 0.0, 1.0);
            uv = in_uv;
            color = in_color;
        }
    ''',
    fragment_shader='''
        #version 330

        out vec4 f_color;
        in vec2 uv;
        in vec4 color;
        uniform sampler2D tex;

        void main() {
            f_color = color * texture(tex, uv);
        }
    ''',
)
QUAD = np.array([0, 1, 2, 0, 2, 3], dtype='i4')


class SpriteArray:
    """Vertex array object to hold textured quads."""
    PROGRAM = TEXTURED_QUADS_PROGRAM

    def __init__(self, ctx, prog, tex, sprites):
        self.tex = tex
        self.ctx = ctx
        self.prog = prog
        self.sprites = list(sprites)
        self._allocate()

    def _allocate(self):
        self.allocated = len(self.sprites)

        # Allocate extra slots in the arrays for faster additions
        extra = max(32 - self.allocated, self.allocated // 2)

        for i, s in enumerate(self.sprites):
            s.array = self
            s.offset = i
            if s.verts is None:
                s._update()

        self.indexes = np.vstack([
            QUAD + 4 * i
            for i in range(self.allocated + extra)
        ])
        self.uvs = np.vstack(
            [s.uvs for s in self.sprites]
            + [np.zeros((4 * extra, 2), dtype='f4')]
        )
        self.verts = np.vstack(
            [s.verts for s in self.sprites]
            + [np.zeros((4 * extra, 7), dtype='f4')]
        )

        self.vbo = self.ctx.buffer(self.verts, dynamic=True)
        self.uvbo = self.ctx.buffer(self.uvs)
        self.ibuf = self.ctx.buffer(self.indexes)
        self.vao = self.ctx.vertex_array(
            self.prog,
            [
                (self.vbo, '3f 4f', 'in_vert', 'in_color'),
                (self.uvbo, '2f', 'in_uv'),
            ],
            self.ibuf
        )

    def add(self, s):
        """Add a sprite to the array.

        If there's unallocated space in the VBO we append the sprite.

        Otherwise we allocate new VBOs.
        """
        s.array = self
        if not s.verts:
            s._update()
        size = len(self.verts) // 4
        if self.allocated < size:
            i = self.allocated
            self.allocated += 1
            self.verts[i * 4:i * 4 + 4] = s.verts
            self.uvs[i * 4:i * 4 + 4] = s.uvs
            self.sprites.append(s)
            s.offset = i

            # TODO: We can send less data with write_chunks()
            self.vbo.write(self.verts)
            self.uvbo.write(self.uvs)
        else:
            self.sprites.append(s)
            self._allocate()

    def delete(self, s):
        """Remove a sprite from the array.

        To do this without resizing the buffer we move a sprite from the
        end of the array into the gap. This means that draw order changes.

        """
        assert s.array is self
        i = s.offset
        j = self.allocated - 1
        self.allocated -= 1
        if i == j:
            self.sprites.pop()
        else:
            moved = self.sprites[i] = self.sprites[j]
            self.sprites.pop()
            moved.offset = i
            self.verts[i * 4:i * 4 + 4] = self.verts[j * 4:j * 4 + 4]
            self.uvs[i * 4:i * 4 + 4] = self.uvs[j * 4:j * 4 + 4]
            # TODO: write only once per frame no matter how many adds/deletes
            self.vbo.write(self.verts)
            self.uvbo.write(self.uvs)
        s.array = None

    def render(self):
        """Render all sprites in the array."""
        self.prog['tex'].value = 0
        self.tex.use(0)
        dirty = False
        for i, s in enumerate(self.sprites):
            if s._dirty:
                self.verts[i * 4:i * 4 + 4] = s.verts
                self.uvs[i * 4:i * 4 + 4] = s.uvs
                s._dirty = False
                dirty = True
        assert self.verts.dtype == 'f4', \
            f"Dtype of verts is {self.verts.dtype}"
        if dirty:
            self.vbo.write(self.verts)
            self.uvbo.write(self.uvs)
        self.vao.render(vertices=self.allocated * 6)


class Sprite(Colorable, Transformable):
    def __init__(
            self,
            layer,
            image,
            anchor=None):
        super().__init__()
        self.verts = None
        self.layer = layer
        self.array = None
        self._image = None
        self._vert_color = np.ones((4, 4), dtype='f4')
        self.image = image  # trigger sprite load and migration

    @property
    def image(self):
        """Get the name of the image for this sprite."""
        return self._image

    @image.setter
    def image(self, name):
        """Set the image."""
        if name == self._image:
            return

        self._image = name
        tex, uvs, vs = self.layer.group.atlas.get(name)
        self.uvs = np.copy(uvs)
        self.orig_verts = np.copy(vs)
        xs = self.orig_verts[:, 0]
        ys = self.orig_verts[:, 1]
        self.width = np.fabs(np.min(xs) - np.max(xs))
        self.height = np.fabs(np.min(ys) - np.max(ys))
        # TODO: apply anchor to the verts
        self._dirty = True

        if not self.array:
            # initial migration into an array
            self.layer._migrate_sprite(self, tex)
        elif tex != self.array.tex:
            # migrate to a different vao
            self.array.delete(self)
            self.layer._migrate_sprite(self, tex)

    def delete(self):
        """Delete this sprite."""
        self.layer._dirty.discard(self)
        self.layer.objects.discard(self)
        self.array.delete(self)
        self.layer = None

    def _set_dirty(self):
        self.layer._dirty.add(self)
        self.verts = None

    bounds = Bounds('self.orig_verts[:, :2]')

    def _update(self):
        xform = self._scale @ self._rot @ self._xlate

        self._vert_color[:] = self._color
        self.verts = np.hstack([
            self.orig_verts @ xform,
            self._vert_color,
        ])
        self._dirty = True



"""An effect where we keep trails from previous frames."""
from typing import Tuple, List
from dataclasses import dataclass

import moderngl

from .base import PostprocessPass


FADE_PROG = """ \
#version 330 core

out vec4 f_color;

uniform float fade;
uniform float dt;

void main()
{
    f_color = vec4(0, 0, 0, 1.0 - pow(fade, dt));
}

"""


COMPOSITE_PROG = """ \
#version 330 core

in vec2 uv;
out vec4 f_color;

uniform sampler2D fb;

void main()
{
    f_color = texture(fb, uv);
}

"""


@dataclass
class Trails:
    """A trails effect."""
    ctx: moderngl.Context
    shadermgr: 'wasabi2d.layers.ShaderManager'
    fade: float = 0.9

    camera: 'wasabi2d.scene.Camera' = None
    _pass: PostprocessPass = None
    _fb: moderngl.Framebuffer = None

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera
        self._fb = camera._make_fb('f2')
        self._fade_pass = PostprocessPass(
            self.ctx,
            self.shadermgr,
            FADE_PROG,
            send_uvs=False
        )
        self._composite_pass = PostprocessPass(
            self.ctx,
            self.shadermgr,
            COMPOSITE_PROG,
        )

    def enter(self, t, dt):
        self._fb.use()
        self._fade_pass.render(
            fade=self.fade,
            dt=dt
        )

    def exit(self, t, dt):
        self.ctx.screen.use()
        self._composite_pass.render(fb=self._fb)



import numpy as np
from pyrr import vector3, matrix33

from ..color import convert_color
from ..allocators.vertlists import VAO
from ..rect import ZRect


Z = vector3.create_unit_length_z()


def identity():
    """Return an identity transformation matrix."""
    return np.identity(3, dtype='f4')


def bounds_from_verts(vertices: np.ndarray) -> ZRect:
    """Given an array of vertices, get the bounds of the object."""
    # Ensure we only have 2D vertices
    vertices = vertices[:, :2]

    # Calculate the bounds
    mins = np.min(vertices, axis=0)
    maxs = np.max(vertices, axis=0)
    return ZRect(mins, maxs - mins)


class Bounds:
    """Descriptor for a property that returns bounds based on vertices.

    We compile an expression to access the vertices in order to be faster and
    save code.

    """

    def __init__(self, expr="self.lst.vertbuf['in_vert']"):
        self.getter = eval(f"lambda self: {expr}")

    def __get__(self, inst, cls=None):
        return bounds_from_verts(self.getter(inst))


class Transformable:
    _angle = 0

    def __init__(self):
        super().__init__()
        self._scale = identity()
        self._rot = identity()
        self._xlate = identity()

    @property
    def pos(self):
        return self._xlate[2][:2]

    @pos.setter
    def pos(self, v):
        assert len(v) == 2
        self._xlate[2][:2] = v
        self._set_dirty()

    @property
    def x(self):
        return self._xlate[2, 0]

    @x.setter
    def x(self, v):
        self._xlate[2, 0] = v
        self._set_dirty()

    @property
    def y(self):
        return self._xlate[2, 1]

    @y.setter
    def y(self, v):
        self._xlate[2, 1] = v
        self._set_dirty()

    @property
    def scale(self):
        p = np.product(np.diagonal(self._scale))
        return np.copysign(np.sqrt(abs(p)), p)

    @scale.setter
    def scale(self, v):
        self._scale[0, 0] = self._scale[1, 1] = v
        self._set_dirty()

    @property
    def angle(self):
        return self._angle

    @angle.setter
    def angle(self, theta):
        assert isinstance(theta, (int, float))
        self._rot = matrix33.create_from_axis_rotation(Z, theta, dtype='f4')
        self._angle = theta
        self._set_dirty()


class Colorable:
    """Mix-in for a primitive that has a 4-component color value."""

    def __init__(self):
        super().__init__()
        self._color = np.ones(4, dtype='f4')

    @property
    def color(self):
        return tuple(self._color)

    @color.setter
    def color(self, v):
        self._color[:] = convert_color(v)
        self._set_dirty()


class AbstractShape(Colorable, Transformable):
    """Base class for polygonal shapes."""
    _stroke_width = 1.0

    def _migrate_stroke(self, vao: VAO):
        """Migrate the stroke into the given VAO."""
        # TODO: dealloc from an existing VAO
        idxs = self._stroke_indices()
        self.vao = vao
        self.lst = vao.alloc(len(self.orig_verts), len(idxs))
        self.lst.indexbuf[:] = idxs
        self.lst.indexbuf += self.lst.vertoff.start
        self._update()

    def _migrate_fill(self, vao: VAO):
        """Migrate the fill into the given VAO."""
        # TODO: dealloc from an existing VAO
        idxs = self._fill_indices()
        self.vao = vao
        self.lst = vao.alloc(len(self.orig_verts), len(idxs))
        self.lst.indexbuf[:] = idxs
        self.lst.indexbuf += self.lst.vertoff.start
        self._update()

    def _set_dirty(self):
        self.layer._dirty.add(self)

    @property
    def stroke_width(self):
        """Get the stroke width, in pixels."""
        return self._stroke_width

    @stroke_width.setter
    def stroke_width(self, v):
        """Set the stroke width in pixels."""
        self._stroke_width = v
        self._set_dirty()

    bounds = Bounds()

    def _update(self):
        xform = self._scale @ self._rot @ self._xlate

        np.matmul(
            self.orig_verts,
            xform[:, :2],
            self.lst.vertbuf['in_vert']
        )
        self.lst.vertbuf['in_color'] = self._color
        if 'in_linewidth' in self.lst.vertbuf.dtype.fields:
            self.lst.vertbuf['in_linewidth'] = self._stroke_width
        self.lst.dirty = True

    def delete(self):
        """Delete this primitive."""
        self.layer._dirty.discard(self)
        self.layer.objects.discard(self)
        self.vao.free(self.lst)
        self.lst = None
        self.layer = None



import math
from functools import partial

import numpy as np
import moderngl

from ..color import convert_color
from ..allocators.vertlists import VAO
from .polygons import AbstractShape
from ..rect import ZRect


#: Shader for a plain color fill
PLAIN_COLOR = dict(
    vertex_shader='''
        #version 330

        uniform mat4 proj;

        in vec2 in_vert;
        in vec4 in_color;
        out vec4 color;

        void main() {
            gl_Position = proj * vec4(in_vert, 0.0, 1.0);
            color = in_color;
        }
    ''',
    fragment_shader='''
        #version 330

        out vec4 f_color;
        in vec4 color;

        void main() {
            f_color = color;
        }
    ''',
)


WIDE_LINE = dict(
    vertex_shader='''
        #version 330

        in vec2 in_vert;
        in vec4 in_color;
        in float in_linewidth;
        out vec4 g_color;
        out float widths;

        void main() {
            gl_Position = vec4(in_vert, 0.0, 1.0);
            g_color = in_color;
            widths = in_linewidth;
        }
    ''',
    geometry_shader="""
#version 330 core
layout (lines_adjacency) in;
//layout (triangle_strip, max_vertices = 2) out;
layout (triangle_strip, max_vertices = 4) out;

in vec4 g_color[];
in float widths[];
out vec4 color;

const float MITRE_LIMIT = 6.0;

vec2 rot90(vec2 v) {
    return vec2(-v.y, v.x);
}

uniform mat4 proj;


void mitre(vec2 a, vec2 b, vec2 c, float width) {
    vec2 ab = normalize(b - a);
    vec2 bc = normalize(c - b);

    if (length(bc) < 1e-6) {
        bc = ab;
    }

    // across bc
    vec2 xbc = rot90(bc);

    vec2 along = normalize(ab + bc);
    vec2 across_mitre = rot90(along);

    float scale = 1.0 / dot(xbc, across_mitre);

    //This kind of works Ok although it does cause the width to change
    // scale = min(scale, MITRE_LIMIT);  // limit extension of the mitre
    vec2 across = width * across_mitre * scale;

    gl_Position = proj * vec4(b + across, 0.0, 1.0);
    EmitVertex();

    gl_Position = proj * vec4(b - across, 0.0, 1.0);
    EmitVertex();
}


void main() {
    color = g_color[1];

    vec2 a = gl_in[0].gl_Position.xy;
    vec2 b = gl_in[1].gl_Position.xy;
    vec2 c = gl_in[2].gl_Position.xy;
    vec2 d = gl_in[3].gl_Position.xy;

    mitre(a, b, c, widths[1]);
    mitre(b, c, d, widths[2]);

    EndPrimitive();
}
""",
    fragment_shader='''
        #version 330

        out vec4 f_color;
        in vec4 color;

        void main() {
            f_color = color;
        }
    ''',
)


def color_vao(
        mode: int,
        ctx: moderngl.Context,
        shadermgr: 'wasabi2d.layers.ShaderManager') -> VAO:
    """Build a BAO for rendering plain colored objects."""
    return VAO(
        mode=mode,
        ctx=ctx,
        prog=shadermgr.get(**PLAIN_COLOR),
        dtype=np.dtype([
            ('in_vert', '2f4'),
            ('in_color', '4f4'),
        ])
    )


def line_vao(
        ctx: moderngl.Context,
        shadermgr: 'wasabi2d.layers.ShaderManager') -> VAO:
    """Build a VAO for rendering lines."""
    return VAO(
        mode=moderngl.LINE_STRIP_ADJACENCY,
        ctx=ctx,
        prog=shadermgr.get(**WIDE_LINE),
        dtype=np.dtype([
            ('in_vert', '2f4'),
            ('in_color', '4f4'),
            ('in_linewidth', 'f4'),
        ])
    )


#: Construct a VAO for rendering the fill of shapes
shape_vao = partial(color_vao, moderngl.TRIANGLES)


class Circle(AbstractShape):
    """A circle drawn with lines."""

    __slots__ = (
        'layer', 'segments',
        '_radius',
        'orig_verts',
    )

    def __init__(
            self,
            layer,
            pos=(0, 0),
            radius=100,
            color=(1, 1, 1, 1),
            stroke_width=1.0,
            segments=None):
        super().__init__()
        self.layer = layer
        self.segments = segments or max(4, round(math.pi * radius))
        self.pos = pos
        self._stroke_width = stroke_width

        # Generate verts now
        theta = np.linspace(0, 2 * np.pi, self.segments).reshape((-1, 1))
        self.base_verts = np.hstack([
            np.cos(theta),
            np.sin(theta),
        ]).astype('f4')

        # There's a duplicate vertex so move the first vertex to the center
        self.base_verts[0][:2] = 0

        # Placeholder for scaled verts
        self.orig_verts = np.ones((self.segments, 3), dtype='f4')

        # Assigning radius generates self.orig_verts
        self.radius = radius

        self._color = convert_color(color)
        self._set_dirty()

    @property
    def radius(self):
        """Get the radius of this circle."""
        return self._radius

    @radius.setter
    def radius(self, r):
        """Set the radius of the circle; rebuild the vertices now."""
        self._radius = r

        np.multiply(
            self.base_verts,
            r,
            self.orig_verts[:, :2]
        )

        self._set_dirty()

    def _stroke_indices(self):
        """Indexes for drawing the stroke as a LINE_STRIP."""
        n = self.segments
        return np.array(
            [n - 1, *range(1, n), 1, 2],
            dtype='i4'
        )

    def _fill_indices(self):
        """Indexes for drawing the fill as TRIANGLES."""
        idxs = np.array([
            (0, i, i + 1)
            for i in range(1, self.segments)
        ], dtype='i4')
        idxs[-1][2] = 1
        return idxs.reshape((-1))

    @property
    def bounds(self):
        x, y = self.pos
        r = self.radius * self.scale
        return ZRect(
            (x - r, y - r),
            (r * 2, r * 2),
        )



"""Separable Gaussian blur."""
from typing import Tuple, List
from dataclasses import dataclass

import moderngl
import numpy as np

from .base import PostprocessPass


def gaussian(x, mu, sig):
    """Calculate a gaussian function."""
    return np.exp(-np.power(x - mu, 2.) / (2 * np.power(sig, 2.)))


# Shader code adapted from https://learnopengl.com/Advanced-Lighting/Bloom
# First pass, blur vertically
BLUR_PROG = """ \
#version 330 core

in vec2 uv;
out vec4 f_color;

uniform sampler2D image;
uniform sampler2D gauss_tex;
uniform float radius;
uniform vec2 blur_direction;


float gauss(float off) {
    return texture(gauss_tex, vec2(off / radius, 0)).r;
}


void main()
{
    vec2 tex_offset = 1.0 / textureSize(image, 0); // gets size of single texel
    vec4 result = texture(image, uv); // current fragment's contribution

    vec2 lookup_stride = tex_offset * blur_direction;
    float weight_sum = 1.0;
    float weight;
    int irad = int(ceil(radius));
    for(int i = 1; i < irad; ++i)
    {
        weight = gauss(i);
        weight_sum += weight * 2;
        result += texture(image, uv + lookup_stride * i) * weight;
        result += texture(image, uv - lookup_stride * i) * weight;
    }
    f_color = result / weight_sum;
}

"""


@dataclass
class Blur:
    """A light bloom effect."""
    ctx: moderngl.Context
    shadermgr: 'wasabi2d.layers.ShaderManager'
    radius: float = 10.0

    camera: 'wasabi2d.scene.Camera' = None
    _blur: PostprocessPass = None
    _fb1: moderngl.Framebuffer = None
    _fb2: moderngl.Framebuffer = None

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera
        self._fb1, self._fb2 = camera._get_temporary_fbs(2, 'f2')
        gauss = gaussian(np.arange(256), 0, 90).astype('f4')
        self._gauss_tex = self.ctx.texture((256, 1), 1, data=gauss, dtype='f4')
        self._blur = PostprocessPass(
            self.ctx,
            self.shadermgr,
            BLUR_PROG
        )

    def enter(self, t, dt):
        self._fb1.use()
        self._fb1.clear()

    def exit(self, t, dt):
        self._fb2.use()
        self._fb2.clear()
        self._blur.render(
            image=self._fb1,
            blur_direction=(0, 1),
            radius=self.radius,
            gauss_tex=self._gauss_tex,
        )
        self.ctx.screen.use()

        self._blur.render(
            image=self._fb2,
            blur_direction=(1, 0),
            radius=self.radius,
            gauss_tex=self._gauss_tex,
        )



"""A distortion of the scene inwards/outwards."""
from dataclasses import dataclass

import moderngl

from .base import PostprocessPass


DISTORT_PROG = """ \
#version 330 core

in vec2 uv;
out vec4 f_color;

uniform sampler2D fb;

uniform float factor;

const vec2 center = vec2(0.5, 0.5);

void main()
{
    vec2 off = uv - center;

    float dist = pow(length(off) * 2, factor) / 2;

    off = dist * normalize(off);

    f_color = texture(fb, center + off);
}

"""


@dataclass
class Punch:
    """A distortion of the scene inwards/outwards."""
    ctx: moderngl.Context
    shadermgr: 'wasabi2d.layers.ShaderManager'
    factor: float = 0.9

    camera: 'wasabi2d.scene.Camera' = None
    _pass: PostprocessPass = None
    _fb: moderngl.Framebuffer = None

    def _set_camera(self, camera: 'wasabi2d.scene.Camera'):
        """Resize the effect for this viewport."""
        self.camera = camera
        self._fb, = camera._get_temporary_fbs(1, 'f2')
        self._pass = PostprocessPass(
            self.ctx,
            self.shadermgr,
            DISTORT_PROG,
        )

    def enter(self, t, dt):
        self._fb.use()
        self._fb.clear()

    def exit(self, t, dt):
        self.ctx.screen.use()
        self._pass.render(
            fb=self._fb,
            factor=self.factor,
        )



"""Example of the punch effect."""
import math
from wasabi2d import run, Scene, event, keys


scene = Scene()

logo = scene.layers[0].add_sprite(
    'wasabi2d',
    pos=(scene.width / 2, scene.height / 2),
)
effect = scene.layers[0].set_effect(
    'blur',
    radius=10
)


@event
def update(t):
    effect.radius = 20 * math.sin(t) ** 2


@event
def on_key_down(key, mod):
    if key == keys.F12:
        scene.screenshot()


run()



"""Base functionality for post-processing effects."""
import numpy as np
import moderngl


POSTPROCESS_VERT_PROGRAM = '''
    #version 330

    in vec2 in_vert;
    in vec2 in_uv;
    out vec2 uv;

    void main() {
        gl_Position = vec4(in_vert.xy, 0.0, 1.0);
        uv = in_uv;
    }
'''


class PostprocessPass:
    """A full-screen postprocessing effect pass.

    This will render a full-screen quad using a single fragment shader.

    """
    # TODO: let the shadermanager itself manage these, so that they
    # are shared between layers etc

    QUAD_INDICES = np.array([0, 1, 2, 0, 2, 3], dtype='i4')
    QUAD_VERTS_UVS = np.array([
        [-1, -1, 0, 0],
        [1, -1, 1, 0],
        [1, 1, 1, 1],
        [-1, 1, 0, 1],
    ], dtype='f4')

    def __init__(
            self,
            ctx: moderngl.Context,
            shadermgr: 'wasabi2d.layers.ShaderManager',
            fragment_shader: str,
            send_uvs: bool = True):
        self.ctx = ctx
        self.prog = shadermgr.get(
            vertex_shader=POSTPROCESS_VERT_PROGRAM,
            fragment_shader=fragment_shader,
        )

        indices = ctx.buffer(self.QUAD_INDICES)
        vs_uvs = ctx.buffer(self.QUAD_VERTS_UVS)

        if send_uvs:
            attribs = (vs_uvs, '2f4 2f4', 'in_vert', 'in_uv')
        else:
            attribs = (vs_uvs, '2f4 8x', 'in_vert')
        self.vao = ctx.vertex_array(
            self.prog,
            [attribs],
            index_buffer=indices
        )

    def render(self, **uniforms):
        """Assign the given uniforms and then render."""
        texnum = 0
        for k, v in uniforms.items():
            if isinstance(v, moderngl.Framebuffer):
                v = v.color_attachments[0]

            if isinstance(v, moderngl.Texture):
                v.use(texnum)
                v = texnum
                texnum += 1

            self.prog[k].value = v
        self.vao.render(moderngl.TRIANGLES, 6)




import io
import os.path
from setuptools import setup, find_packages


path = os.path.join(os.path.dirname(__file__), 'README.md')
with io.open(path, encoding='utf8') as f:
    LONG_DESCRIPTION = f.read()


setup(
    name='wasabi2d',
    version='1.0.0',
    description="A convenient 2D OpenGL games framework",
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',
    url='https://github.com/lordmauve/wasabi2d',
    author='Daniel Pope',
    author_email='mauve@mauveweb.co.uk',
    packages=find_packages(include='wasabi2d*'),
    package_data={'wasabi2d': ['data/*']},
    install_requires=open('requirements.txt').read().splitlines(),
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Education',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3.6',
        'Topic :: Education',
        'Topic :: Games/Entertainment',
    ],
    test_suite='tests'
)



"""Example of the light bloom effect."""
from wasabi2d import run, Scene, event, keys, Vector2
from wasabi2d.actor import Actor


scene = Scene()

logo = Actor(
    scene.layers[0].add_sprite(
        'wasabi2d',
        pos=(scene.width / 2, scene.height / 2),
    )
)
logo.v = Vector2(100, -100)
scene.layers[0].set_effect(
    'trails',
    fade=0.7
)


@event
def update(dt):
    logo.pos += logo.v * dt
    if logo.top < 0 or logo.bottom >= scene.height:
        logo.v.y *= -1
    if logo.left < 0 or logo.right >= scene.width:
        logo.v.x *= -1


@event
def on_key_down(key, mod):
    if key == keys.F12:
        scene.screenshot()


run()



"""Wrapper around window creation."""
import sys
import math
import numpy as np
import pygame
import pygame.image
import pygame.transform
import pygame.display
import moderngl
from pyrr import Matrix44

from . import clock
from .layers import LayerGroup
from .loaders import set_root


class Scene:
    """Top-level interface for renderable objects."""

    def __init__(
            self,
            width=800,
            height=600,
            antialias=0,
            title="wasabi2d",
            rootdir=None):
        self.width = width
        self.height = height

        self._recording = False

        if rootdir is None:
            rootdir = sys._getframe(1).f_globals['__file__']
        set_root(rootdir)

        pygame.init()

        glconfig = {
            'GL_CONTEXT_MAJOR_VERSION': 4,
            'GL_CONTEXT_MINOR_VERSION': 3,
            'GL_CONTEXT_PROFILE_MASK': pygame.GL_CONTEXT_PROFILE_CORE,
        }

        if antialias:
            glconfig.update({
                'GL_MULTISAMPLEBUFFERS': 1,
                'GL_MULTISAMPLESAMPLES': antialias,
            })

        for k, v in glconfig.items():
            k = getattr(pygame, k)
            pygame.display.gl_set_attribute(k, v)

        self.screen = pygame.display.set_mode(
            (width, height),
            flags=pygame.OPENGL | pygame.DOUBLEBUF,
            depth=24
        )
        self.title = title
        ctx = self.ctx = moderngl.create_context()

        self.camera = Camera(ctx, width, height)
        self.layers = LayerGroup(ctx, self.camera)

        ctx.enable(moderngl.BLEND)
        ctx.blend_func = moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA

        from . import event
        event(self.draw)

        self.background = (0.0, 0.0, 0.0)

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, title):
        self._title = title
        pygame.display.set_caption(title)

    def screenshot(self, filename=None):
        """Take a screenshot.

        If filename is not given, save to a file named screenshot_<time>.png
        in the current directory.

        """
        import datetime
        if filename is None:
            now = datetime.datetime.now()
            filename = f'screenshot_{now:%Y-%m-%d_%H:%M:%S.%f}.png'
        data = self.ctx.screen.read(components=3)
        assert len(data) == (self.width * self.height * 3), \
            f"Received {len(data)}, expected {self.width * self.height * 3}"
        img = pygame.image.fromstring(data, (self.width, self.height), 'RGB')
        img = pygame.transform.flip(img, False, True)
        pygame.image.save(img, filename)

    def record_video(self, filename=None):
        """Start recording a video.

        Video will be encoded in MPEG4 format.

        This requires an ffmpeg binary to be located on $PATH.

        If filename is not given, save to a file named video_<time>.mp4
        in the current directory.
        """
        import subprocess
        import datetime
        if not filename:
            now = datetime.datetime.now()
            filename = f'video_{now:%Y-%m-%d_%H:%M:%S.%f}.mp4'
        self._recording = filename
        command = [
            'ffmpeg',
            '-y',  # (optional) overwrite output file if it exists
            '-f', 'rawvideo',
            '-vcodec', 'rawvideo',
            '-s', f'{self.width}x{self.height}',  # size of one frame
            '-pix_fmt', 'rgb24',
            '-r', '60',  # frames per second
            '-i', '-',  # The imput comes from a pipe
            '-vf', 'vflip',

            # These options are needed for uploads to Twitter
            '-vcodec', 'libx264',
            '-pix_fmt', 'yuv420p',
            '-strict', '-2',
            '-an',  # Tells FFMPEG not to expect any audio
            filename,
        ]
        self._ffmpeg = subprocess.Popen(
            command,
            stdin=subprocess.PIPE,
            bufsize=10**8
        )
        print("Recording video...")
        from . import event
        event.lock_fps = True

    def stop_recording(self):
        """Finish recording the current video."""
        self._ffmpeg.stdin.close()
        ret = self._ffmpeg.wait()
        if ret == 0:
            print("Saved recording to", self._recording)
        else:
            print("Error writing video.")
        self._recording = None
        from . import event
        event.lock_fps = False

    def toggle_recording(self) -> bool:
        """Start or stop recording video.

        Return True if recording started.
        """
        if not self._recording:
            self.record_video()
            return True
        else:
            self.stop_recording()

    def _vid_frame(self):
        data = self.ctx.screen.read(components=3)
        self._ffmpeg.stdin.write(data)
        self._ffmpeg.stdin.flush()

    def draw(self, t, dt):
        assert len(self.background) == 3, \
            "Scene.background must be a 3-element tuple."
        if self._recording:
            self._vid_frame()
        self.ctx.clear(*self.background)
        self.layers.render(self.camera.proj, t, dt)


class Camera:
    """The camera/viewport through which the scene is viewed.

    As well as constructing a projection matrix for the screen, and offering
    some dynamic effects such as screen shake, a Camera serves as a manager
    for temporary framebuffers corresponding to the viewport size.
    """

    def __init__(self, ctx, width, height):
        self.ctx = ctx
        self.width = width
        self.height = height
        hw = self.width * 0.5
        hh = self.height * 0.5
        self._proj = Matrix44.orthogonal_projection(
            left=-hw,
            right=hw,
            top=hh,
            bottom=-hh,
            near=-1000,
            far=1000
        ).astype('f4')
        self._xform = np.identity(4, dtype='f4')
        self._cam_offset = np.zeros(2, dtype='f4')
        self._cam_vel = np.zeros(2, dtype='f4')
        self._pos = np.zeros(2, dtype='f4')
        self._fbs = {}
        self.pos = hw, hh

    def _get_temporary_fbs(self, num=1, dtype='f1'):
        """Get temporary framebuffer objects of the given dtype."""
        temps = self._fbs.setdefault(dtype, [])
        while len(temps) < num:
            fb = self._make_fb(dtype)
            temps.append(fb)
        return temps[:num]

    def _make_fb(self, dtype='f1', div_x=1, div_y=1):
        """Make a new framebuffer corresponding to this viewport."""
        tex = self.ctx.texture(
            (self.width // div_x, self.height // div_y),
            4,
            dtype=dtype
        )
        tex.repeat_x = tex.repeat_y = False
        return self.ctx.framebuffer([tex])

    @property
    def pos(self):
        return -self._xform[-1][:2]

    @pos.setter
    def pos(self, v):
        assert len(v) == 2
        self._pos[:] = v
        self._xform[-1][:2] = self._cam_offset - self._pos

    @property
    def proj(self):
        return self._xform @ self._proj

    def screen_shake(self, dist=25):
        """Trigger a screen shake effect.

        The camera will be offset from ``.pos`` by ``dist`` in a random
        direction; then steady itself in a damped harmonic motion.

        """
        theta = np.random.uniform(0, math.tau)
        basis = np.array([theta, + math.pi * 0.5])
        self._cam_offset[:] = dist * np.sin(basis)
        self._xform[-1][:2] = self._cam_offset - self._pos
        clock.schedule_unique(self._steady_cam, 0.01)

    def _steady_cam(self):
        dt = 0.05  # guarantee stable behaviour
        self._cam_offset += self._cam_vel * dt
        self._cam_vel -= self._cam_offset * (300 * dt)
        self._cam_vel *= 0.1 ** dt
        self._cam_offset *= 0.01 ** dt
        if np.sum(self._cam_vel ** 2) < 1e-3 \
                and np.sum(self._cam_offset ** 2) < 1e-2:
            self._cam_offset[:] = self._cam_vel[:] = 0
        else:
            clock.schedule_unique(self._steady_cam, 0.01)
        self._xform[-1][:2] = self._cam_offset - self._pos

